#%% modules

import copy
import numpy as np
import numpy.ma as ma
import pandas as pd
import cv2
import os
import pprint
import sys

from math import inf

from phenopype import _vars
from phenopype import utils_lowlevel
from phenopype.core import preprocessing
from phenopype.core import segmentation
from phenopype.core import visualization

#%% classes


class motion_tracker(object):
    """
    Initialize motion tracker class; extract information (length, fps, codec, 
    etc.) from input video and pass to other tracking methods.

    Parameters
    ----------
    video_path : str
        path to video
    at_frame : int, optional
        frame index to be used to extract the video information
        
    """

    def __init__(self, video_path, at_frame=1):

        ## extract frame
        if os.path.isfile(video_path):
            capture = cv2.VideoCapture(video_path)
            idx = 0
            while capture.isOpened():
                idx += 1
                if idx == at_frame:
                    ret, frame = capture.read()
                    break
                else:
                    capture.grab()
                if cv2.waitKey(1) & 0xFF == 27:
                    break
        else:
            print("No compatible video file found under provided path")
            return

        ## properties
        self.path = video_path
        self.name = os.path.basename(self.path)
        self.nframes = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))
        self.fps = capture.get(cv2.CAP_PROP_FPS)
        self.fourcc_str = utils_lowlevel._decode_fourcc(
            capture.get(cv2.CAP_PROP_FOURCC)
        )
        self.length = str(
            str(int((self.nframes / self.fps) / 60)).zfill(2)
            + ":"
            + str(
                int(
                    (
                        ((self.nframes / self.fps) / 60)
                        - int((self.nframes / self.fps) / 60)
                    )
                    * 60
                )
            ).zfill(2)
        )
        self.dimensions = tuple(reversed(frame.shape[0:2]))
        if frame.shape[2] == 3:
            self.is_colour = True

        ## for masks
        self.image = frame
        self.df_image_data = pd.DataFrame(
            {
                "filename": self.name,
                "filepath": self.path,
                "filetype": self.fourcc_str,
                "width": self.dimensions[0],
                "height": self.dimensions[1],
            },
            index=[0],
        )

        ## default
        self.flag_save_video = False

        ## release capture
        capture.release()

        ## feeddback
        print("\n")
        print("--------------------------------------------------------------")
        print('Input video properties - "' + self.name + '":\n')
        print(
            "Frames per second: "
            + str(self.fps)
            + "\nN frames: "
            + str(self.nframes)
            + "\nLength: "
            + str(self.length)
            + " (mm:ss)"
            + "\nDimensions: "
            + str(self.dimensions)
            + "\nColour video: "
            + str(self.is_colour)
            + "\nFourCC code: "
            + str(self.fourcc_str)
        )
        print("--------------------------------------------------------------")

    def video_output(
        self,
        video_format=None,
        save_suffix="out",
        dirpath=None,
        fps=None,
        save_colour=None,
        dimensions=None,
        resize=1,
    ):
        """
        Set properties of output video file. Most settings can be left blank, 
        so that settings from the input video will be applied
        
        Parameters
        ----------
        video_format: str, optional
            format of the output video file. needs to be a fourcc-string
            https://www.fourcc.org/codecs.php
        save_suffix: str, optional
            name for the output video file (defaut: "input-filename" + _out.avi)
        dirpath: str, optional
            save directory for the output video file. will be created if not existing
        fps: int, optional
            frames per second of the output video file
        dimensions: tuple, optional
            dimensions (width, length) of the output video file
        resize: int, optional
            factor by which to resize the output video dimensions
        save_colour: bool, optional
            should the saved video frames be in colour 
        """

        self.flag_save_video = True

        ## name and path
        name = (
            os.path.splitext(self.name)[0]
            + "_"
            + save_suffix
            + os.path.splitext(self.name)[1]
        )
        if dirpath.__class__.__name__ == "NoneType":
            dirpath = os.path.dirname(self.path)
        else:
            if not os.path.isdir(dirpath):
                q = input("Save folder {} does not exist - create?.".format(dirpath))
                if q in ["True", "true", "y", "yes"]:
                    os.makedirs(dirpath)
                else:
                    print("Directory not created - aborting")
                    return
        path_out = os.path.join(dirpath, name)

        ## video properties
        if video_format.__class__.__name__ == "NoneType":
            fourcc_str = self.fourcc_str
        else:
            fourcc_str = video_format
        fourcc_out = cv2.VideoWriter_fourcc(*fourcc_str)
        if fps.__class__.__name__ == "NoneType":
            fps_out = self.fps
        if dimensions.__class__.__name__ == "NoneType":
            dimensions_out = self.dimensions
        if save_colour.__class__.__name__ == "NoneType":
            self.flag_save_colour = self.is_colour
        else:
            self.flag_save_colour = save_colour

        self.resize_factor = resize
        if self.resize_factor < 1:
            res_msg = " (original resized by factor " + str(self.resize_factor) + ")"
        else:
            res_msg = ""
        dimensions_out = (
            int(dimensions_out[0] * self.resize_factor),
            int(dimensions_out[1] * self.resize_factor),
        )

        ## start video-writer
        self.writer = cv2.VideoWriter(
            path_out, fourcc_out, fps_out, dimensions_out, self.flag_save_colour
        )

        print("\n")
        print("--------------------------------------------------------------")
        print('Output video settings - "' + self.name + '":\n')
        print(
            "Save name: "
            + name
            + "\nSave dir: "
            + os.path.abspath(dirpath)
            + "\nFrames per second: "
            + str(fps_out)
            + "\nDimensions: "
            + str(dimensions_out)
            + res_msg
            + "\nColour video: "
            + str(self.flag_save_colour)
            + "\nFormat (FourCC code): "
            + fourcc_str
        )
        print("--------------------------------------------------------------")

    def detection_settings(
        self,
        skip=5,
        warmup=0,
        start_after=0,
        finish_after=0,
        history=60,
        threshold=10,
        detect_shadows=True,
        mode="MOG",
        methods=None,
        masks=None,
        c_mask=False,
        c_mask_shape="rect",
        c_mask_size=50,
    ):

        """
        Set properties of output video file. Most settings can be left at their 
        default value.
        
        Parameters
        ----------
        skip: int, optional
            how many frames to skip between each capture
        warmup: int, optional
            warmup period in seconds for the background subtractor
        start_after: int, optional
            start after X seconds
        finish_after: int, optional
            finish after X seconds
        history: int, optional
            how many frames to use for fg-bg subtraction algorithm
        threshold: int, optional
            sensitivity-level for fg-bg subtraction algorithm (lower = more 
            sensitive)
        detect_shadows: bool, optional
            attempt to detect shadows - will be returned as gray pixels
        mode: {"MOG", "KNN"} str, optional
            type of fg-bg subtraction algorithm
        methods: method or list of methods, optional
            list with tracking_method objects
        c_mask: bool, optional 
            consecutive masking. if multiple methods are defined, the objects 
            detected first will mask the objects detected in subsequent methods
        c_mask_shape: {"rect", "ellipse", "contour"} str, optional
            which shape should the consecutive mask have
        c_mask_size: int, optional
            area in pixels that is added around the mask
            
        """
        ## kwargs
        self.skip = skip
        self.warmup = warmup
        self.start = start_after
        self.finish = finish_after
        self.flag_detect_shadows = detect_shadows
        self.flag_consecutive = c_mask
        self.consecutive_shape = c_mask_shape
        self.consecutive_size = c_mask_size
        self.masks = masks

        ## select background subtractor
        if mode == "MOG":
            self.fgbg_subtractor = cv2.createBackgroundSubtractorMOG2(
                int(history * (self.fps / self.skip)),
                threshold,
                self.flag_detect_shadows,
            )
        elif mode == "KNN":
            self.fgbg_subtractor = cv2.createBackgroundSubtractorKNN(
                int(history * (self.fps / self.skip)),
                threshold,
                self.flag_detect_shadows,
            )

        ## return settings of methods
        if not methods.__class__.__name__ == "NoneType":
            if methods.__class__.__name__ == "tracking_method":
                methods = [methods]

            self.methods = methods
            for m in self.methods:
                print("-----------------------------")
                print("Detection method \"{}\":\n".format(m.label))
                m._print_settings()
                print("-----------------------------")
                print("\n")

        print("\n")
        print("--------------------------------------------------------------")
        print('Motion detection settings - "' + self.name + '":\n')
        print(
            "Background-subtractor: "
            + str(mode)
            + "\nHistory: "
            + str(history)
            + " seconds"
            + "\nSensitivity: "
            + str(threshold)
            + "\nRead every nth frame: "
            + str(self.skip)
            + "\nDetect shadows: "
            + str(self.flag_detect_shadows)
            + "\nStart after n seconds: "
            + str(self.start)
            + "\nFinish after n seconds: "
            + str(self.finish if self.finish > 0 else " - ")
        )
        print("--------------------------------------------------------------")

    def run_tracking(self, feedback=True, canvas="overlay", overlay_weight=0.5, **kwargs):
        """
        Start motion tracking procedure. Enable or disable video feedback, 
        output and select canvas (overlay of detected objects, foreground mask,
        input video).
        
        Parameters
        ----------
        feedback: bool, optional
            show output of tracking
        canvas: {"overlay","fgmask","input"} str, optional
            the background for the ouput video
        overlay_weight: float (default: 0.5)
            if canvas="overlay", how transparent should the overlay should be

        """

        ## kwargs
        flag_feedback = feedback

        ## check if settings have been called
        if not hasattr(self, "fgbg_subtractor"):
            self.detection_settings()

        ## initialize
        self.df = pd.DataFrame()
        self.idx1, self.idx2 = (0, 0)
        self.capture = cv2.VideoCapture(self.path)
        self.start_frame = int(self.start * self.fps)
        if self.finish > 0:
            self.finish_frame = int(self.finish * self.fps)
        else:
            self.finish_frame = self.nframes

        if all(hasattr(self, attr) for attr in ["methods", "masks"]):
            for m in self.methods:
                m._apply_masks(frame=self.image, masks=self.masks)

        ## loop thrpugh frames
        while self.capture.isOpened():

            ## frame indexing
            self.idx1, self.idx2 = (self.idx1 + 1, self.idx2 + 1)
            if self.idx2 == self.skip:
                self.idx2 = 0

            ## time conversion
            mins = str(int((self.idx1 / self.fps) / 60)).zfill(2)
            secs = str(int((((self.idx1 / self.fps) / 60) - int(mins)) * 60)).zfill(2)
            self.time_stamp = (
                "Time: "
                + mins
                + ":"
                + secs
                + "/"
                + self.length
                + " - Frames: "
                + str(self.idx1)
                + "/"
                + str(int(self.nframes))
            )

            ## end-of-frames-control
            if self.idx1 == self.finish_frame - 1:
                self.capture.release()
                if self.flag_save_video:
                    self.writer.release()
                break

            ## capture frame
            if (
                self.idx1 > self.start_frame - int(self.warmup * self.fps)
                and self.idx2 == 0
            ):
                self.ret, self.frame = self.capture.read()

                ## skip empty frames
                if self.ret == False:
                    continue
                else:
                    capture_frame = True
                    if (
                        self.idx1 < self.start_frame
                        and self.idx1 > self.start_frame - int(self.warmup * self.fps)
                    ):
                        print(self.time_stamp + " - warmup")
                    else:
                        print(self.time_stamp + " - captured")
            else:
                self.capture.grab()
                print(self.time_stamp)
                continue

            ## if captured, apply masks > apply methods > write to output
            if capture_frame == True:

                # initiate tracking
                fgmask = self.fgbg_subtractor.apply(self.frame)
                fgmask_copy = copy.deepcopy(fgmask)
                self.frame_overlay = self.frame

                # apply methods
                if "methods" in vars(self):
                    idx = 0
                    for m in self.methods:
                        fgmask_copy = copy.deepcopy(fgmask)
                        (
                            self.fgmask_mod,
                            self.overlay,
                            self.method_contours,
                            self.frame_df,
                        ) = m._run(frame=self.frame, fgmask=fgmask_copy)
                        idx += 1

                        # shadowing of methods
                        if self.flag_consecutive and idx < len(self.methods):
                            self.method_mask = np.zeros_like(fgmask_copy)
                            for contour in self.method_contours:
                                if self.consecutive_shape == "contour":
                                    self.method_mask = cv2.drawContours(
                                        self.method_mask,
                                        [contour],
                                        0,
                                        utils_lowlevel._get_bgr("white"),
                                        -1,
                                    )  # Draw filled contour in mask
                                elif self.consecutive_shape == "ellipse":
                                    self.method_mask = cv2.ellipse(
                                        self.method_mask,
                                        cv2.fitEllipse(contour),
                                        utils_lowlevel._get_bgr("white"),
                                        -1,
                                    )
                                elif self.consecutive_shape in ["rect", "rectangle"]:
                                    rx, ry, rw, rh = cv2.boundingRect(contour)
                                    cv2.rectangle(
                                        self.method_mask,
                                        (int(rx), int(ry)),
                                        (int(rx + rw), int(ry + rh)),
                                        utils_lowlevel._get_bgr("white"),
                                        -1,
                                    )
                                kernel = cv2.getStructuringElement(
                                    cv2.MORPH_RECT,
                                    (self.consecutive_size, self.consecutive_size),
                                )
                                self.method_mask = cv2.dilate(
                                    self.method_mask, kernel, iterations=1
                                )
                            fgmask = cv2.subtract(fgmask, self.method_mask)

                        # create overlay for each method
                        self.frame_overlay = cv2.addWeighted(
                            self.frame_overlay, 1, self.overlay, overlay_weight, 0
                        )

                        # make data.frame
                        self.frame_df.insert(0, "frame_abs", self.idx1)
                        self.frame_df.insert(
                            1, "frame", int((self.idx1 - self.start_frame) / self.skip)
                        )
                        self.frame_df.insert(2, "mins", mins)
                        self.frame_df.insert(3, "secs", secs)
                        # self.df = self.df.append(
                        #     self.frame_df, ignore_index=True, sort=False
                        # )
                        self.df = pd.concat([self.df, self.frame_df], ignore_index=True, sort=False)


                ## select canvas
                if "methods" in vars(self):
                    if canvas == "overlay":
                        self.canvas = self.frame_overlay
                    elif canvas == "fgmask":
                        self.canvas = fgmask_copy
                    elif canvas == "fgmask_mod":
                        self.canvas = self.fgmask_mod
                    else:
                        self.canvas = self.frame
                else:
                    self.canvas = fgmask_copy

                ## resize
                if "resize_factor" in vars(self):
                    self.canvas = cv2.resize(
                        self.canvas,
                        (0, 0),
                        fx=self.resize_factor,
                        fy=self.resize_factor,
                    )

                ## convert to colour
                if len(self.canvas.shape) < 3:
                    self.canvas = cv2.cvtColor(self.canvas, cv2.COLOR_GRAY2BGR)

                ## draw masks
                if not self.masks.__class__.__name__ == "NoneType":
                    for key, value in self.masks[_vars._mask_type].items():
                        self.canvas = visualization.draw_mask(
                            self.canvas, {_vars._mask_type: {key: value}}, label=True
                        )

                ## feedback
                if flag_feedback == True:
                    cv2.namedWindow("phenopype", cv2.WINDOW_AUTOSIZE)
                    cv2.imshow("phenopype", self.canvas)

                # save output
                if self.flag_save_video == True:
                    if self.flag_save_colour and len(self.canvas.shape) < 3:
                        self.canvas = cv2.cvtColor(self.canvas, cv2.COLOR_GRAY2BGR)
                    self.writer.write(self.canvas)

            ## keep stream open
            if cv2.waitKey(1) & 0xFF == 27:
                self.capture.release()
                self.writer.release()
                break

        ## cleanup
        self.capture.release()
        if "writer" in vars(self):
            self.writer.release()
        cv2.destroyAllWindows()

        ## return DataFrame
        debug =  kwargs.get("debug", False)
        
        if debug:
            return self
        else:
            return self.df


class tracking_method:

    """
    Constructs a tracking method that can be supplied to the motion_tracker 
    class.
    
    Parameters
    ----------
    label: str, optional
        label for all objects detected by this method
    blur: int, optional
        blurring of fgbg-mask (kernel size)
    threshold: int, optional
        binarization of fgbg-mask after blurring (threshold value)
    remove_shadows: bool, optional
        if motion_tracker has detect_shadows=True, they can be removed here
    min_area : int, optional
        minimum contour area in pixels to be included
    max_area : int, optional
        maximum contour area in pixels to be included
    min_length : int, optional
        minimum diameter of boundary circle to be included
    max_length : int, optional
        maximum diameter of boundary circle to be included
    mode: {"single", "multiple"} str, optional
        track "multiple", or "single" (biggest by diameter) objects
    remove_shadows: bool, optional
        remove shadows if shadow-detection is actived in MOG-algorithm
    overlay_colour: {"red", "green", "blue", "black", "white"} str, optional
        which colour should tracked objects have
    operations: list (default: ["length", "area"])
        determines the type of operations to be performed on the detected objects:
            - "diameter" of the bounding circle of our object
            - "area" within the contour of our object
            - "grayscale" mean and standard deviation of grayscale pixel values 
              inside the object contours
            - "grayscale_background" background within boundingbox of contour
            - "bgr" mean and standard deviation of blue, green and red pixel 
              values inside the object contours
    """

    def __init__(
        self,
        label="m1",
        blur=5,
        threshold=127,
        remove_shadows=True,
        min_length=0,
        max_length=inf,
        min_area=0,
        max_area=inf,
        mode="multiple",
        overlay_colour="red",
        operations=[],
    ):

        ## kwargs
        self.blur_kernel = blur
        self.label = label
        self.overlay_colour = utils_lowlevel._get_bgr(overlay_colour)
        self.min_length, self.max_length = min_length, max_length
        self.min_area, self.max_area = min_area, max_area
        self.mode = mode
        self.operations = operations
        self.threshold_value = threshold
        self.remove_shadows = remove_shadows

    def _apply_masks(self, frame, masks):
        """
        Applies masks drawn using the motion_tracker.
        
        Internal reference - don't call this directly. 
        """
        if masks.__class__.__name__ == "dict":
            if _vars._mask_type in masks:
                self.mask_bool = {}
                for key, value in masks[_vars._mask_type].items():
                    polygons = masks[_vars._mask_type][key]["data"][
                        _vars._mask_type
                    ]
                    label = masks[_vars._mask_type][key]["data"]["label"]
                    for coords in polygons:
                        mask_bool = utils_lowlevel._create_mask_bool(frame, coords)
                        self.mask_bool[label] = mask_bool

    def _print_settings(self, width=30, indent=1, compact=True):
        """
        Prints the settings of the tracking method. 
        
        Internal reference - don't call this directly. 
        """

        print_dict = {}
        for key, value in self.__dict__.items():
            if type(value) in [str, float, int, tuple, bool]:
                print_dict[key] = value
                
        pretty = pprint.PrettyPrinter(width=width, compact=compact, indent=indent)
        pretty.pprint(print_dict)

    def _run(self, frame, fgmask):
        """
        Run tracking method on current frame. 
        
        Internal reference - don't call this directly.      
        """

        ## initialize
        self.overlay = np.zeros_like(frame)
        self.overlay_bin = np.zeros(frame.shape[0:2], dtype=np.uint8)
        self.frame_df = pd.DataFrame()

        if self.remove_shadows == True:
            ret, fgmask = cv2.threshold(fgmask, 128, 255, cv2.THRESH_BINARY)

        ## blur
        fgmask = preprocessing.blur(fgmask, self.blur_kernel)

        # ## threshold
        fgmask = segmentation.threshold(
            fgmask, method="binary", invert=True, value=self.threshold_value
        )

        ## find contours
        contours, hierarchy = cv2.findContours(
            fgmask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )

        ## perform operations on contours
        if len(contours) > 0:
            list_contours, list_area, list_length, list_center_coordinates = (
                [],
                [],
                [],
                [],
            )
            df_list, df_column_names = [], []

            # check if contour matches min/max length provided
            for contour in contours:
                if contour.shape[0] > 4:
                    center, radius = cv2.minEnclosingCircle(contour)
                    length = int(radius * 2)
                    area = int(cv2.contourArea(contour))
                    if all(
                        [
                            length > self.min_length and length < self.max_length,
                            area > self.min_area and area < self.max_area,
                        ]
                    ):
                        list_length.append(length)
                        list_area.append(area)
                        list_contours.append(contour)
                        list_center_coordinates.append(center)

            if len(list_contours) > 0:
                # if single biggest contour:
                if self.mode == "single":
                    if len(contours) == 1:
                        pass
                    elif len(contours) > 1:
                        max_idx = np.argmax(list_length)
                        list_contours = [list_contours[max_idx]]
                        list_length = [list_length[max_idx]]
                        list_area = [list_area[max_idx]]
                        list_center_coordinates = [list_center_coordinates[max_idx]]

                list_x, list_y = [], []
                list_grayscale, list_grayscale_background = [], []
                list_b, list_g, list_r = [], [], []
                list_mask_check = []

                for contour, center in zip(list_contours, list_center_coordinates):

                    # operations
                    x = int(center[0])
                    y = int(center[1])
                    list_x.append(x)
                    list_y.append(y)

                    if "mask_bool" in vars(self):
                        temp_list = []
                        for key, val in self.mask_bool.items():
                            temp_list.append(val[y, x])
                        list_mask_check.append(temp_list)

                    rx, ry, rw, rh = cv2.boundingRect(contour)
                    frame_roi = frame[ry : ry + rh, rx : rx + rw]
                    frame_roi_gray = cv2.cvtColor(frame_roi, cv2.COLOR_BGR2GRAY)
                    mask_roi = fgmask[ry : ry + rh, rx : rx + rw]

                    if any("grayscale" in o for o in self.operations):
                        grayscale = ma.array(
                            data=frame_roi_gray, mask=np.logical_not(mask_roi)
                        )
                        list_grayscale.append(int(np.mean(grayscale)))

                    if any("grayscale_background" in o for o in self.operations):
                        grayscale_background = ma.array(
                            data=frame_roi_gray, mask=mask_roi
                        )
                        if not grayscale_background.mask.all():
                            list_grayscale_background.append(
                                int(np.mean(grayscale_background))
                            )
                        else:
                            list_grayscale_background.append(9999)

                    if any("bgr" in o for o in self.operations):
                        b = ma.array(
                            data=frame_roi[:, :, 0], mask=np.logical_not(mask_roi)
                        )
                        list_b.append(int(np.mean(b)))
                        g = ma.array(
                            data=frame_roi[:, :, 1], mask=np.logical_not(mask_roi)
                        )
                        list_g.append(int(np.mean(g)))
                        r = ma.array(
                            data=frame_roi[:, :, 2], mask=np.logical_not(mask_roi)
                        )
                        list_r.append(int(np.mean(r)))

                    # drawing
                    self.overlay = cv2.drawContours(
                        self.overlay, [contour], 0, self.overlay_colour, -1
                    )  # Draw filled contour in mask
                    self.overlay = cv2.putText(
                        self.overlay,
                        self.label,
                        (x, y),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        1,
                        self.overlay_colour,
                        1,
                        cv2.LINE_AA,
                    )
                    self.overlay = cv2.rectangle(
                        self.overlay,
                        (rx, ry),
                        (int(rx + rw), int(ry + rh)),
                        self.overlay_colour,
                        2,
                    )

                df_list = df_list + [list_x]
                df_list = df_list + [list_y]
                df_column_names = df_column_names + ["x", "y"]

                if any("diameter" in o for o in self.operations):
                    df_list = df_list + [list_length]
                    df_column_names.append("diameter")

                if any("area" in o for o in self.operations):
                    df_list = df_list + [list_area]
                    df_column_names.append("area")

                if any("grayscale" in o for o in self.operations):
                    df_list = df_list + [list_grayscale]
                    df_column_names.append("grayscale")

                if any("grayscale_background" in o for o in self.operations):
                    df_list = df_list + [list_grayscale_background]
                    df_column_names.append("grayscale_background")

                if any("bgr" in o for o in self.operations):
                    df_list = df_list + [list_b]
                    df_list = df_list + [list_g]
                    df_list = df_list + [list_r]
                    df_column_names = df_column_names + ["b", "g", "r"]

                frame_df = pd.DataFrame(data=df_list)
                frame_df = frame_df.transpose()
                frame_df.columns = df_column_names
                frame_df["label"] = self.label

                if "mask_bool" in vars(self):

                    mask_df = pd.DataFrame(list_mask_check, columns=[*self.mask_bool])
                    self.frame_df = pd.concat(
                        [frame_df.reset_index(drop=True), mask_df], axis=1
                    )

                else:
                    self.frame_df = frame_df

                self.contours = list_contours

                return fgmask, self.overlay, self.contours, self.frame_df

            else:
                frame_df = pd.DataFrame()
                return fgmask, self.overlay, [], self.frame_df

        else:
            frame_df = pd.DataFrame()
            return fgmask, self.overlay, [], self.frame_df



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

# -- Project information -----------------------------------------------------

project = 'phenopype'
copyright = '2021, Moritz Lürig'
author = 'Moritz Lürig'

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinx.ext.githubpages',
    'sphinx.ext.autosectionlabel',
    'recommonmark',
    'nbsphinx',
    'sphinx_design',
]

autodoc_member_order = 'bysource'
suppress_warnings = ['autosectionlabel.*']

# The master toctree document.
master_doc = 'tutorials'

# error handling
nbsphinx_allow_errors = True

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['.ipynb_checkpoints/*.ipynb']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'bootstrap'
html_theme = "furo"
html_show_sourcelink = True
html_last_updated_fmt = '%b %d, %Y, %X'
html_title = "phenopype docs"
html_static_path = ['_assets/css']
html_css_files = ['custom.css']
templates_path = ["_templates"]
# html_additional_pages = {
    # "index": "lander.html"
# }
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {'**': ['localtoc.html', 'sourcelink.html', 'searchbox.html']}
# html_sidebars = {
    # '**': ['localtoc.html'],
    # 'index': []
# }

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'phenopypedoc'



#%% modules

import pytest
import numpy as np

import phenopype as pp

#%% tests

def test_blur(image):
    
    image_blurred = pp.preprocessing.blur(image)
    image_blurred = pp.preprocessing.blur(image, kernel_size=4)
    image_blurred = pp.preprocessing.blur(image, method="gaussian")
    image_blurred = pp.preprocessing.blur(image, method="median")
    image_blurred = pp.preprocessing.blur(image, method="bilateral")

    assert not (image == image_blurred).all()

def test_create_mask(image, mask_polygon):
    
    annotations = mask_polygon

    annotations = pp.preprocessing.create_mask(
        image, 
        annotations=annotations, 
        annotation_id="a", 
        interactive=False,
        )

    assert len(annotations) > 0
    
    
def test_detect_mask(image):
    
    annotations = pp.preprocessing.detect_mask(
        image, 
        circle_args={"param1": 150, "param2": 150, "min_radius":1000}, 
        resize=0.5
        )
    
    annotations = pp.preprocessing.detect_mask(
        image, 
        circle_args={"param1": 150, "param2": 150, "max_radius":150}, 
        resize=0.5
        )

    assert len(annotations) > 0
    
def test_create_reference(reference_created, settings):
    
    image = pp.load_image(pytest.reference_image_path)
    
    annotations = reference_created
    
    annotations = pp.preprocessing.create_reference(
        image,
        annotations=annotations, 
        interactive=False
        )
    
    assert len(annotations) > 0
    
def test_detect_reference(image, reference_created, settings):
    
    annotations = reference_created
    
    reference_image = pp.load_image(pytest.reference_image_path)
    
    coords = annotations["reference"]["a"]["data"]["mask"][0]
    template = reference_image[coords[0][1]:coords[2][1], coords[0][0]:coords[1][0]]
    px_ratio = annotations["reference"]["a"]["data"]["reference"][0]
    unit = annotations["reference"]["a"]["data"]["reference"][1]
    
    black_dummy = np.zeros((5001, 5001, 3), dtype=np.uint8)
    
    annotations = pp.preprocessing.detect_reference(
        image=black_dummy,
        template=template,
        correct_colours=True,
        px_ratio=px_ratio,
        unit=unit
        )
    
    annotations = pp.preprocessing.detect_reference(
        image=image,
        template=template,
        resize=0.25,
        correct_colours=True,
        px_ratio=px_ratio,
        unit=unit
        )
        
    assert len(annotations) > 0

    
def test_decompose_image(image):
    
    mod = pp.preprocessing.decompose_image(image, "bier")
    mod = pp.preprocessing.decompose_image(image, "raw")
    mod = pp.preprocessing.decompose_image(image, "hsv")
    mod = pp.preprocessing.decompose_image(image, "hsv", channels=0)
    mod = pp.preprocessing.decompose_image(image, "hsv", channels=[1,2])
    mod = pp.preprocessing.decompose_image(image, channel="blue")
    mod = pp.preprocessing.decompose_image(image, "rgb", channel="blue")

    canvas = pp.visualization.select_canvas(mod, multi_channel=True)

    assert not (image == canvas).all()
    
    
def test_write_comment(image, comment):
    
    annotations = comment
    
    annotations = pp.preprocessing.write_comment(
        image=image,
        annotations=annotations,
        interactive=False,
        )

    assert len(annotations) > 0






#%% modules
import os
import mock
import pytest

import phenopype as pp

#%% tests

def test_project(settings):
    with mock.patch('builtins.input', return_value='n'):
        project = pp.Project(root_dir=pytest.project_root_dir_1)
    with mock.patch('builtins.input', return_value='y'):
        project = pp.Project(root_dir=pytest.project_root_dir_1)
    with mock.patch('builtins.input', return_value='n'):
        project = pp.Project(root_dir=pytest.project_root_dir_1, overwrite=True)
    with mock.patch('builtins.input', return_value='y'):
        project = pp.Project(root_dir=pytest.project_root_dir_1, overwrite=True)
        
    project = pp.Project(root_dir=pytest.project_root_dir_1)
        
    assert os.path.isdir(project.root_dir)    
    
    
        
def test_project_add_files(project, settings):
    
    project.add_files(
        image_dir=pytest.image_dir, 
        mode="link", 
        include="stickle"
        )
    
    project.add_files(
        image_dir=pytest.image_dir, 
        include="stickle",
        resize_factor=0.5,
        overwrite=True
        )
    
    project.add_files(
        image_dir=pytest.image_dir, 
        mode="copy", 
        include="stickle",
        overwrite=True
        )
    
    assert len(project.dir_paths) > 0
    
    
    
def test_project_add_config(project, settings):
       
    project.add_config(
        tag=pytest.tag_1, 
        template_path="wrong-template"
        )
    
    project.add_config(
        tag=pytest.tag_1, 
        template_path=pytest.template_path_1
        )

    config = pp.utils_lowlevel._load_yaml(
        os.path.join(
            project.root_dir, 
            project.dir_paths[0], 
            "pype_config_" + pytest.tag_1 + ".yaml"
            ))
    
    assert config["config_info"]["template_name"] == os.path.basename(pytest.template_path_1)
    
    

def test_project_add_reference_template(project, reference_created, mask_polygon, settings):
    
    annotations = {**reference_created, **mask_polygon}
    
    project.add_reference_template(
        image_path=pytest.reference_image_path,
        reference_id="ref1",
        feedback=False,
        overwrite=True,
        annotations=annotations,
        )
    
    assert os.path.isfile(os.path.join(project.root_dir, "reference", "ref1_template.tif"))
    
    
    
def test_project_edit_config(project, settings):
    
    with mock.patch('builtins.input', return_value="y"):
        project.edit_config(
            tag=pytest.tag_1, 
            target=pytest.edit_config_target, 
            replacement=pytest.edit_config_replacement
            )
        
    path = os.path.join(project.dir_paths[0],  "pype_config_" + pytest.tag_1 + ".yaml")
    
    with open(path) as f:
        if pytest.edit_config_replacement in f.read():
            success = True
        else:
            success = False
            
    assert success
    





#%% modules

from collections import defaultdict
import copy
import json
import os
import sys

import cv2
import numpy as np
import pandas as pd

from phenopype import __version__
from phenopype import _vars
from phenopype import config
from phenopype import decorators
from phenopype import utils
from phenopype import utils_lowlevel as ul
from phenopype.core import preprocessing

#%% functions

def convert_annotation(
    annotations,
    annotation_type,
    annotation_id,
    annotation_type_new,
    annotation_id_new,
    overwrite=False,
    **kwargs
):
    """
    convert coordinates from one annotation type to another. currently, only
    converting from contour to mask format is possible

    Parameters
    ----------

    annotations : dict
        A phenopype annotation dictionary.
    annotation_type : str | list of str
        If dict contains multiple annotation types, select one or more to
        load. None will load all types.
    annotation_id : str | list of str
        If file contains multiple annotation IDs, select one or more to
        load. None will load all IDs within a type.
    annotation_type_new : str
       target annotation type
    annotation_id_new : str | list of str, optional
       target annotation id
    overwrite : bool, optional
        if target exists, overwrite? The default is False.

    Returns
    -------
    None.


    """

    # =============================================================================
    # annotation management

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
        prep_msg="- extracting annotation:",
    )

    # =============================================================================
    # setup

    ## convert_annotation
    fun_name = sys._getframe().f_code.co_name

    # =============================================================================
    # method

    if annotation_type_new == _vars._mask_type:

        ## method setup
        if "label" in annotation["data"]:
            label = annotation["data"]["label"]
        else:
            label = kwargs.get("label", "mask1")
        include = kwargs.get("include", True)

        annotation_old_data = annotation["data"][annotation_type]

        annotation_new_data = []
        for idx, coord_list in enumerate(annotation_old_data):
            annotation_new_data.append(ul._convert_arr_tup_list(coord_list,  add_first=True))


        annotation = {
            "info": {
                "annotation_type": annotation_type_new,
                "phenopype_function": fun_name,
                "phenopype_version": __version__,
            },
            "settings": {
                "tool": "polygon",
            },
            "data": {
                "label": label,
                "include": include,
                "n": len(annotation_new_data),
                annotation_type_new: annotation_new_data,
            },
        }

    return ul._update_annotations(
        annotations=annotations,
        annotation=annotation,
        annotation_type=annotation_type_new,
        annotation_id=annotation_id_new,
        kwargs=kwargs,
    )




def export_csv(
    annotations,
    dir_path,
    annotation_type=None,
    image_name=None,
    overwrite=False,
    **kwargs
):
    """
    export annotations from json to csv format.

    Parameters
    ----------
    annotations : dict
        A phenopype annotation dictionary.
    dir_path : str
        To which folder should the csv file be exported.
        PYPE: Automatically set to the current image directory
    annotation_type : str / list, optional
        Which annotation types should be exported - can be string or list of strings.
    image_name : str
       Image name to be added as a column
       PYPE: Automatically adds the image name.
    overwrite : bool, optional
        Should an existing csv file be overwritten. The default is False.


    Returns
    -------
    None.

    """

    ## dirpath
    if not image_name:
        ul._print(
            "Warning: missing image_name argument - exported CSV will not contain information about source-image",
            lvl=1
        )

    ## file name formatting
    if kwargs.get("save_prefix"):
        save_prefix = kwargs.get("save_prefix") + "_"
    else:
        save_prefix = ""

    if kwargs.get("save_suffix"):
        save_suffix = "_" + kwargs.get("save_suffix")
    else:
        save_suffix = ""

    ## annotations copy
    annotations = copy.deepcopy(annotations)

    ## format annotation type
    if annotation_type.__class__.__name__ == "NoneType":
        ul._print("- no annotation_type selected - exporting all annotations")
        annotation_types = list(annotations.keys())
    elif annotation_type.__class__.__name__ == "str":
        annotation_types = [annotation_type]
    elif annotation_type.__class__.__name__ in ["list", "CommentedSeq"]:
        annotation_types = annotation_type

    # ## dry run
    # annotation_types1 = []
    # for annotation_type in annotation_types:
    #     if len(annotations[annotation_type].keys()) > 0:
    #         annotation_types1.append(annotation_type)
    # annotation_types = annotation_types1

    for annotation_type in annotation_types:

        list_flattened = []

        ## comment
        if annotation_type == _vars._comment_type and annotation_type in annotations.keys():
            for annotation_id in annotations[annotation_type].keys():

                label = annotations[annotation_type][annotation_id]["data"]["label"]
                comment = annotations[annotation_type][annotation_id]["data"][annotation_type]
                list_flattened.append(
                    pd.DataFrame(
                        {
                            **{"image_name": image_name},
                            **{"annotation_type": annotation_type},
                            **{"annotation_id": annotation_id},
                            **{label:comment},
                        },
                        index=[0],
                    )
                )

        ## contour
        if annotation_type == _vars._contour_type:
            for annotation_id in annotations[annotation_type].keys():
                for idx, (coords, support) in enumerate(zip(
                    annotations[annotation_type][annotation_id]["data"][annotation_type],
                    annotations[annotation_type][annotation_id]["data"]["support"],
                ), 1):
                    list_flattened.append(
                        # df_temp = pd.DataFrame(ul._convert_arr_tup_list(coords)[0], columns=["x","y"])
                        pd.DataFrame(
                            {
                                **{"image_name": image_name},
                                **{"annotation_type": annotation_type},
                                **{"annotation_id": annotation_id},
                                **{"contour_idx": idx},
                                **{"center_x": support["center"][0]},
                                **{"center_y": support["center"][1]},
                                **{"area": support["area"]},
                                **{"diameter": support["diameter"]},
                                **{"hierarchy_level": support["hierarchy_level"]},
                                **{"hierarchy_idx_child": support["hierarchy_idx_child"]},
                                **{"hierarchy_idx_parent": support["hierarchy_idx_parent"]},
                            },
                            index=[0],
                        )
                    )


        ## landmark
        if annotation_type == _vars._landmark_type:
            for annotation_id in annotations[annotation_type].keys():
                lm_tuple_list = list(zip(*annotations[annotation_type][annotation_id]["data"][annotation_type]))
                list_flattened.append(
                    pd.DataFrame.from_dict(
                        {
                            **{"image_name": image_name},
                            **{"annotation_type": annotation_type},
                            **{"annotation_id": annotation_id},
                            **{"landmark_idx": range(1,len(lm_tuple_list[0])+1)},
                            **{"x_coords": lm_tuple_list[0]},
                            **{"y_coords": lm_tuple_list[1]},

                        },
                    )
                )

        ## polyline
        if annotation_type == _vars._line_type:
            for annotation_id in annotations[annotation_type].keys():
                for idx, (coords, length) in enumerate(zip(
                    annotations[annotation_type][annotation_id]["data"][annotation_type],
                    annotations[annotation_type][annotation_id]["data"]["lengths"],
                ), 1):
                    # line_tuple_list = list(zip(*coords))
                    list_flattened.append(
                        pd.DataFrame(
                            {
                                **{"image_name": image_name},
                                **{"annotation_type": annotation_type},
                                **{"annotation_id": annotation_id},
                                **{"line_idx": idx},
                                # **{"node_idx": range(1,len(line_tuple_list[0])+1)},
                                # **{"x_coords": line_tuple_list[0]},
                                # **{"y_coords": line_tuple_list[1]},
                                **{"length": length},
                            },
                            index=[0],
                        )
                    )

        ## reference
        if annotation_type == _vars._reference_type:
            for annotation_id in annotations[annotation_type].keys():
                distance = annotations[annotation_type][annotation_id]["data"][annotation_type][0]
                unit = annotations[annotation_type][annotation_id]["data"][annotation_type][1]
                list_flattened.append(
                    pd.DataFrame(
                        {
                            **{"image_name": image_name},
                            **{"annotation_type": annotation_type},
                            **{"annotation_id": annotation_id},
                            **{"distance": distance},
                            **{"unit": unit},
                        },
                        index=[0],
                    )
                )


        ## shape_features
        if annotation_type == _vars._shape_feature_type:
            for annotation_id in annotations[annotation_type].keys():
                contour_id = annotations[annotation_type][annotation_id]["settings"][_vars._contour_type + "_id"]
                for idx, annotation in enumerate(
                    annotations[annotation_type][annotation_id]["data"][annotation_type], 1
                ):
                    list_flattened.append(
                        pd.DataFrame(
                            {
                                **{"image_name": image_name},
                                **{"annotation_type": annotation_type},
                                **{"annotation_id": annotation_id},
                                **{"contour_id": contour_id},
                                **{"contour_idx": idx},
                                **annotation,
                            },
                            index=[0],
                        )
                    )

        ## texture_features
        if annotation_type == _vars._texture_feature_type:
            for annotation_id in annotations[annotation_type].keys():
                contour_id = annotations[annotation_type][annotation_id]["settings"][_vars._contour_type + "_id"]
                for idx, annotation in enumerate(
                    annotations[annotation_type][annotation_id]["data"][annotation_type], 1
                ):
                    list_flattened.append(
                        pd.DataFrame(
                            {
                                **{"image_name": image_name},
                                **{"annotation_type": annotation_type},
                                **{"annotation_id": annotation_id},
                                **{"contour_id": contour_id},
                                **{"contour_idx": idx},
                                **annotation,
                            },
                            index=[0],
                        )
                    )

        if len(list_flattened) > 0:
            ul._print("- exported csv for type " + annotation_type)
            df = pd.concat(list_flattened)
            filepath = os.path.join(
                dir_path, save_prefix + annotation_type + save_suffix + ".csv"
            )
            df.to_csv(filepath, index=False)


def load_annotation(filepath, annotation_type=None, annotation_id=None, tag=None,**kwargs):
    """
    load phenopype annotations file

    Parameters
    ----------
    filepath : str
        Path to JSON file containing annotations
    annotation_type : str | list of str, optional
        If file contains multiple annotation types, select one or more to
        load. None will load all types. The default is None.
    annotation_id : str | list of str, optional
        If file contains multiple annotation IDs, select one or more to
        load. None will load all IDs within a type. The default is None.

    Returns
    -------
    dict
        Loaded annotations.

    """

    ## load annotation file
    if os.path.isfile(filepath):
        with open(filepath) as file:
            try:
                annotation_file = json.load(file)
            except Exception as ex:
                print(
                    "load_annotation: " + str(ex.__class__.__name__) + " - " + str(ex)
                )
                return
        annotation_file = defaultdict(dict, annotation_file)

    elif os.path.isdir(filepath):
        if tag.__class__.__name__ == "NoneType":
            print("Attempting to load directory without specifying tag - aborting")
            return
        else:
            filepath = os.path.join(filepath, "annotations_{}.json".format(tag))
            with open(filepath) as file:
                annotation_file = json.load(file)
    else:
        print("Annotation file not found")
        return

    ## parse serialized array
    for annotation_type1 in annotation_file:
        for annotation_id1 in annotation_file[annotation_type1]:
            for section in annotation_file[annotation_type1][annotation_id1]:
                for key, value in annotation_file[annotation_type1][annotation_id1][
                    section
                ].items():
                    if key in [
                        x
                        for x in _vars._annotation_types
                        if not x in [_vars._comment_type, _vars._reference_type,]
                    ] + ["support"]:
                        if type(value) == str:
                            value = eval(value)
                        if key == _vars._contour_type:
                            value = [np.asarray(elem, dtype=np.int32) for elem in value]
                        elif annotation_type1 == _vars._landmark_type:
                            value = [tuple(elem) for elem in value]
                    annotation_file[annotation_type1][annotation_id1][section][
                        key
                    ] = value

    ## subsetting
    while True:

        ## filter by annotation type
        if annotation_type.__class__.__name__ == "NoneType":
            ul._print("- no annotation_type selected - returning all annotations")
            annotation = annotation_file
            break
        elif annotation_type.__class__.__name__ == "str":
            annotation_type_list = [annotation_type]
        elif annotation_type.__class__.__name__ == "list":
            annotation_type_list = annotation_type
            pass
        annotation_subset = defaultdict(dict)
        for annotation_type in annotation_type_list:
            annotation_subset[annotation_type] = annotation_file[annotation_type]

        ## filter by annotation id
        if annotation_id.__class__.__name__ == "NoneType":
            ul._print(
                "- no annotation_id selected - returning all annotations of type: {}".format(
                    annotation_type_list
                )
            )
            annotation = annotation_subset
            break
        elif annotation_id.__class__.__name__ == "str":
            annotation_id_list = [annotation_id]
        elif annotation_id.__class__.__name__ == "list":
            annotation_id_list = annotation_id
            pass
        annotation = defaultdict(dict)
        for annotation_type in annotation_type_list:
            for annotation_id in annotation_id_list:
                if annotation_id in annotation_subset[annotation_type]:
                    annotation[annotation_type][annotation_id] = annotation_subset[
                        annotation_type
                    ][annotation_id]
        break

    ## return
    return dict(annotation)



def save_annotation(
    annotations,
    dir_path,
    annotation_type=None,
    annotation_id=None,
    overwrite=False,
    **kwargs
):
    """
    save phenopype annotations file

    Parameters
    ----------
    annotation : dict
        Annotation dictionary formatted by phenopype specifications:

        .. code-block:: python

            {
                annotation_type = {
                    annotation_id = annotation,
                    annotation_id = annotation,
                    ...
                    },
                annotation_type = {
                    annotation_id = annotation,
                    annotation_id = annotation,
                    ...
                    },
                ...
            }

    annotation_id : str, optional
        String ("a"-"z") specifying the annotation ID to be saved. None will
        save all IDs. The default is None.
    dir_path : str, optional
        Path to folder where annotation should be saved. None will save the
        annotation in the current Python working directory. The default is None.
    file_name : str, optional
        Filename for JSON file containing annotation. The default is
        "annotations.json".
    overwrite : bool, optional
        Overwrite options should file or annotation entry in file exist:

            - False = Neither file or entry will be overwritten
            - True or "entry" = A single entry will be overwritten
            - "file" = The whole will be overwritten.

        The default is False.

    Returns
    -------
    None.

    """

    ## kwargs
    indent = kwargs.get("indent", 4)
    file_name = kwargs.get("file_name","annotations")

    if not file_name.endswith(".json"):
        file_name = file_name + ".json"

    ## filepath
    filepath = os.path.join(dir_path, file_name)
    annotations = copy.deepcopy(annotations)

    ## open existing json or create new
    while True:
        if os.path.isfile(filepath):
            if overwrite in [False, True, "entry"]:
                with open(filepath) as file:
                    try:
                        annotation_file = json.load(file)
                        annotation_file = defaultdict(dict, annotation_file)
                    except Exception as ex:
                        ul._print(
                            "load_annotation: "
                            + str(ex.__class__.__name__)
                            + " - "
                            + str(ex)
                        )
                        ul._print("Could not read {} - creating new file.".format(filepath))
                        annotation_file = defaultdict(dict, {})

                ul._print("- loading existing annotation file")
                break
            elif overwrite == "file":
                pass
                ul._print('- overwriting annotation file (overwrite="file")')
        else:
            ul._print("- creating new annotation file")
            pass
        annotation_file = defaultdict(dict)
        break

    annotations = defaultdict(dict, annotations)

    ## subset annotation types
    if annotation_type.__class__.__name__ == "NoneType":
        ul._print("- no annotation_type selected - exporting all annotations")
        annotation_types = list(annotations.keys())
    elif annotation_type.__class__.__name__ == "str":
        annotation_types = [annotation_type]
    elif annotation_type.__class__.__name__ in ["list", "CommentedSeq"]:
        annotation_types = annotation_type

    for annotation_type in annotation_types:

    ## write annotations to output dict
    # for annotation_type in annotations:
        for annotation_id in annotations[annotation_type]:
            annotation_id_new = annotation_id
            if str(annotation_id) in annotation_file[annotation_type]:
                if overwrite in [True, "entry"]:
                    annotation_file[annotation_type][annotation_id_new] = annotations[
                        annotation_type
                    ][annotation_id]
                    ul._print(
                        '- updating annotations of type "{}" with id '
                        '"{}" in "{}" (overwrite="entry")'.format(
                            annotation_type, annotation_id, file_name
                        )
                    )
                else:
                    ul._print(
                        '- annotations of type "{}" with id "{}" already '
                        'exists in "{}" (overwrite=False)'.format(
                            annotation_type, annotation_id, file_name
                        )
                    )
            else:
                annotation_file[annotation_type][annotation_id_new] = annotations[
                    annotation_type
                ][annotation_id]
                ul._print(
                    '- writing annotations of type "{}" with id '
                    '"{}" to "{}"'.format(annotation_type, annotation_id, file_name)
                )

    ## remove indents from annotations arrays and lists
    for annotation_type in annotation_file:
        for annotation_id in annotation_file[annotation_type]:
            for section in annotation_file[annotation_type][annotation_id]:
                for key, value in annotation_file[annotation_type][annotation_id][section].items():

                    ## unindent lists for better legibility
                    if key in [x for x in _vars._annotation_types if x not in [_vars._comment_type, _vars._reference_type]] + ["support"]:
                        if isinstance(value, list) and value and isinstance(value[0], np.ndarray):
                            value = [elem.tolist() for elem in value]
                        value = [ul._NoIndent(elem) for elem in value]
                    elif isinstance(value, (tuple, list)):
                        value = ul._NoIndent(value)
                    annotation_file[annotation_type][annotation_id][section][key] = value

    ## save
    with open(filepath, "w") as file:
        json.dump(
            annotation_file, file, indent=indent, cls=ul._NoIndentEncoder
        )

def save_ROI(
    image,
    annotations,
    dir_path,
    file_name,
    channel="raw",
    counter=True,
    which="all",
    suffix=None,
    rotate=False,
    rotate_padding=5,
    angle_apply = None,
    align="v",
    ext="jpg",
    background="original",
    canvas_dim=False,
    canvas_fit=False,
    min_dim=False,
    max_dim=False,
    padding=True,
    **kwargs
    
):
    """
    
    Save a region of interest (ROI) indicated by contour or mask coordinates as
    a crop of the original image, optionally with white background
    
    Parameters
    ----------
    image : ndarray
        An image containing regions of interest (ROI).
    annotations : dict
        A phenopype annotation dict containing one or more contour coordinate
        entries.
    dir_path : str, optional
        Path to folder where annotation should be saved. None will save the
        annotation in the current Python working directory. The default is None.
    file_name : str
        Name for ROI series (should reflect image content, not "ROI" or the like
        which is specified with prefix or suffix arguments). The contour index
        will be added as a numeric string at the end of the filename.
    channel : str, optional
        Which channel to save. The default is "raw".
    counter : TYPE, optional
        Whether to add a contour to the filename. The default is True.
    prefix : str, optional
        Prefix to prepend to individual ROI filenames. The default is None.
    suffix : str, optional
        Suffix to append to individual ROI filenames. The default is "roi".
    extension : str, optional
        New e. The default is "png".
    background : str, optional
        Sets background. The default is "original", providing the background 
        contained within the bounding rectangle. "transparent" will produce a
        png file with tranparent background. "white", "black" or any other 
        color will produce a different background color.
        
    Returns
    -------
    None.

    """





    # =============================================================================
    # annotation management

    annotation_type = kwargs.get("annotation_type", _vars._mask_type)
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    data = annotation["data"][annotation_type]
    
    # =============================================================================
    # prep save
    
    ## add suffix 
    if suffix is None:
        suffix = ""
    else:
        suffix = "_" + suffix
        
    ## add extension
    if "." not in ext:
        ext = "." + ext
        
    # =============================================================================
    # run
    
    if which == "max":
        area = list()
        for idx, roi_coords in enumerate(data):
    
            if annotation_type == _vars._mask_type:
                coords = ul._convert_tup_list_arr(roi_coords)[0]
            else:
                coords = copy.deepcopy(roi_coords)
        
            area.append(int(cv2.contourArea(coords)))
        
            data = [data[area.index(max(area))]]
    
    if not channel=="raw":
        image = preprocessing.decompose_image(image, channel, **kwargs)
        
    for idx, roi_coords in enumerate(data):

        if annotation_type == _vars._mask_type and not isinstance(roi_coords, np.ndarray):
            coords = ul._convert_tup_list_arr(roi_coords)
        else:
            coords = copy.deepcopy(roi_coords)


        # =============================================================================
        ## create roi and mask
        
        if min_dim:
            (x_center, y_center), diameter = cv2.minEnclosingCircle(coords)
            half_length = min_dim // 2
            rx = int(max(0, x_center - half_length))
            ry = int(max(0, y_center - half_length))
            rh, rw = min_dim, min_dim
            roi = copy.deepcopy(image[ry : ry + rh, rx : rx + rw])
        else:
            rx, ry, rw, rh = cv2.boundingRect(coords)
            roi = copy.deepcopy(image[ry : ry + rh, rx : rx + rw])
            half_length = 0
            
        if roi.ndim == 3:
            nlayer = 3
        else:
            nlayer = 1

        roi_mask = np.zeros(roi.shape[:2], dtype="uint8")
        roi_mask = cv2.drawContours(
            image=roi_mask,
            contours=[coords],
            contourIdx=0,
            thickness=-1,
            color=255,
            offset=(-rx, -ry),
        )
        
        # =============================================================================
        ## rotate roi and mask
        
        if rotate:
            if angle_apply == None:
                angle = ul._get_orientation(coords)
                if align == "h":
                    angle = angle + 90
                elif align == "v":
                    angle = angle - 90
            else:
                angle = angle_apply
            roi_rot = ul._rotate_image(roi, angle, allow_crop=False)
            roi_mask_rot = ul._rotate_image(roi_mask, angle, allow_crop=False)
            contour, _ = cv2.findContours(roi_mask_rot, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
            rx, ry, rw, rh = cv2.boundingRect(contour[0])
            roi = roi_rot[ry : ry + rh, rx : rx + rw]
            roi_mask = roi_mask_rot[ry : ry + rh, rx : rx + rw]
            
            ## pad to avoid aliasing
            roi = cv2.copyMakeBorder(roi, 1,1,1,1, cv2.BORDER_CONSTANT, value=0)    
            roi_mask = cv2.copyMakeBorder(roi_mask, 1,1,1,1, cv2.BORDER_CONSTANT, value=0)  
            
        # =============================================================================
        ## mount on canvas
        
        if canvas_dim:
            
            ## make canvas
            roi_canvas = np.zeros((canvas_dim, canvas_dim, nlayer), dtype=np.uint8)
            roi_mask_canvas = np.zeros((canvas_dim, canvas_dim), dtype=np.uint8)

            # calculate the new top-left coordinates for the smaller ROI
            xpos = (canvas_dim - roi.shape[1]) // 2
            ypos = (canvas_dim - roi.shape[0]) // 2
            
            ## skip if ROI is too large for canvas
            if any([xpos<=0, ypos<=0]):
                if canvas_fit:
                    roi = utils.resize_image(roi, max_dim=canvas_dim-2)
                    roi_mask = utils.resize_image(roi_mask, max_dim=canvas_dim-2)
                    xpos = max((canvas_dim - roi.shape[1]) // 2,1)
                    ypos = max((canvas_dim - roi.shape[0]) // 2,1)
                else:
                    ul._print(f"ROI {idx+1} too big for chosen canvas: {roi.shape[:2]} vs {(canvas_dim,canvas_dim)}. Skipping!", lvl=1)
                    continue
            
            # Place the smaller ROI onto the larger ROI canvas
            roi_canvas[ypos:ypos + roi.shape[0], xpos:xpos + roi.shape[1]] = roi
            roi_mask_canvas[ypos:ypos + roi_mask.shape[0], xpos:xpos + roi_mask.shape[1]] = roi_mask
            
            ## replace original
            roi,roi_mask  = roi_canvas, roi_mask_canvas
            
        # =============================================================================
        ## background handling
        
        if background=="original":           
            pass
        else:
            if background=="transparent":           
                roi_alpha = np.zeros((roi.shape[0], roi.shape[1], 4), dtype=np.uint8)
                roi_alpha[:,:,0:3] = roi
                roi_alpha[:, :, 3] = roi_mask
                roi_canvas = roi_alpha
                ext = ".png"
            else:
                roi[roi_mask==0] = ul._get_bgr(background)
                roi_canvas = roi
            roi = roi_canvas
                
                
        # =============================================================================
        ## resizing final ROI
        
        if max_dim:
            roi = utils.resize_image(roi, max_dim=max_dim)
            
        # =============================================================================
        ## saving          
                
        if not kwargs.get("training_data"):
                             
            ## add counter
            if counter:
                roi_name = file_name + suffix + "_" + str(idx+1).zfill(3) + ext
            else:
                roi_name = file_name + suffix + ext  
            
            
            save_path = os.path.join(dir_path, roi_name)
            saved = cv2.imwrite(save_path, roi) 
            
            if saved:
                ul._print("- saving ROI: {}".format(roi_name))
            else:
                ul._print("- something went wrong - didn't save ROI")
                
        else:
            return roi, roi_mask
        
@decorators.legacy_args
def save_canvas(
        image, 
        file_path, 
        ext="jpg",
        resize=0.5,
        overwrite=True,
        **kwargs,
        ):
    """


    Parameters
    ----------
    image : ndarray
        A canvas to be saved.
    save_suffix : str
        A suffix to be appended to the filename.
    dir_path : str
        Path to directory to save canvas.
    resize: float, optional
        resize factor for the image (1 = 100%, 0.5 = 50%, 0.1 = 10% of

    Returns
    -------
    None.

    """
    
    image = utils.resize_image(image, resize)

    utils.save_image(
        image=image,
        file_path=file_path,
        ext=ext,
        overwrite=overwrite,
        **kwargs
    )



#%% modules

import mock
import pytest
import os

import phenopype as pp

#%% tests


def test_load_yaml():

    y = pp.utils_lowlevel._load_yaml(pytest.template_path_4, legacy=True, typ="safe")
    y = pp.utils_lowlevel._load_yaml(pytest.template_path_1)

    pp.utils_lowlevel._show_yaml(y)




#%% modules

import copy
import cv2
import numpy as np
import sys

from dataclasses import make_dataclass
from math import inf

from phenopype import __version__
from phenopype import _vars
from phenopype import utils
from phenopype import utils_lowlevel as ul
from phenopype.core import preprocessing, visualization


#%% functions


def contour_to_mask(
    annotations,
    include=True,
    label=None,
    box_margin=0,
    largest=True,
    **kwargs,
):

    """
    Converts a contour to a mask annotation, e.g. for the purpose of creating an ROI
    and exporting it or for subsequent segmentation inside that mask. Creates a 
    rectangle bounding box around the largest or all contours. 

    Parameters
    ----------
    annotation: dict
        phenopype annotation containing contours
    include : bool, optional
        include or exclude area inside mask
    label : str, optional
        text label for this mask and all its components
    box_margin : int, optional
        margin that is added between the outer perimeter of the contour and the box
    largest : bool, optional
        either use only the largest contour or concatenate all supplied contours

    Returns
    -------
    annotation: dict
        phenopype annotation containing contours
    """
    ## fun name
    fun_name = sys._getframe().f_code.co_name

    # =============================================================================
    # annotation management

    ## get contours
    annotation_type = _vars._contour_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    contours = annotation["data"][annotation_type]
    contours_support = annotation["data"]["support"]
    
    annotation_type = ul._get_annotation_type(fun_name)
    annotation_id = kwargs.get("annotation_id", None)

    # =============================================================================
    # setup
    
    label = kwargs.get("label")
    
    ## take single largest contour
    if largest:
        area_list = []
        for cnt in contours_support:
            area_list.append(cnt["area"])
            
        contour = contours[area_list.index(max(area_list))]
        
    ## concatenate all contours
    else:
        contour = np.concatenate(contours)
        
    # =============================================================================
    # process

    rx, ry, rw, rh = cv2.boundingRect(contour)
    b = box_margin

    mask_coords = [[
        (rx-b,ry-b),
        (rx+rw+b, ry-b),
        (rx+rw+b, ry+rh+b),
        (rx-b, ry+rh+b),
        (rx-b,ry-b)
        ]]
    
    # =============================================================================
    # assemble results

    annotation = {
        "info": {
            "phenopype_function": fun_name,
            "phenopype_version": __version__,
            "annotation_type": annotation_type,
        },
        "settings": {
        },
        "data": {
            "label": label,
            "include": include,
            "n": 1,
            annotation_type: mask_coords,
            },
    }

    # =============================================================================
    # return

    return ul._update_annotations(
        annotations=annotations,
        annotation=annotation,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

def detect_contour(
    image,
    approximation="simple",
    retrieval="ext",
    match_against=None,
    apply_drawing=False,
    offset_coords=[0, 0],
    stats_mode="circle",
    min_nodes=3,
    max_nodes=inf,
    min_area=0,
    max_area=inf,
    min_diameter=0,
    max_diameter=inf,
    **kwargs,
):

    """
    Find objects in binarized image. The input image needs to be a binarized image
    A flavor of different approximation algorithms and retrieval intstructions can be applied. 
    The contours can also be filtered for minimum area, diameter or nodes
    (= number of corners). 

    Parameters
    ----------
    image : ndarray
        input image (binary)
    approximation : {"none", "simple", "L1", "KCOS"] str, optional
        contour approximation algorithm:
            - none: no approximation, all contour coordinates are returned
            - simple: only the minimum coordinates required are returned
            - L1: Teh-Chin chain approximation algorithm 
            - KCOS: Teh-Chin chain approximation algorithm 
    retrieval : {"ext", "list", "tree", "ccomp", "flood"} str, optional
        contour retrieval procedure:
            - ext: retrieves only the extreme outer contours
            - list: retrieves all of the contours without establishing any 
              hierarchical relationships
            - tree: retrieves all of the contours and reconstructs a full 
              hierarchy of nested contours
            - ccomp: retrieves all of the contours and organizes them into a 
              two-level hierarchy (parent and child)
            - flood: flood-fill algorithm
    offset_coords : tuple, optional
        offset by which every contour point is shifted.
    min_nodes : int, optional
        minimum number of coordinates
    max_nodes : int, optional
        maximum number of coordinates 
    min_area : int, optional
        minimum contour area in pixels
    max_area : int, optional
        maximum contour area in pixels
    min_diameter : int, optional
        minimum diameter of boundary circle
    max_diameter : int, optional
        maximum diameter of boundary circle

    Returns
    -------
    annotation: dict
        phenopype annotation containing contours
    """

    # =============================================================================
    # setup
    
    fun_name = sys._getframe().f_code.co_name
    
    # =============================================================================
    # annotation management

    annotations = kwargs.get("annotations", {})

    ## drawings
    annotation_type = _vars._drawing_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
        prep_msg="- combining contours and drawing:",
    )

    annotation_type = ul._get_annotation_type(fun_name)
    annotation_id = kwargs.get("annotation_id", None)
        
    # =============================================================================
    # setup

    image_bin = copy.deepcopy(image)

    if len(image_bin.shape) > 2:
        ul._print("Multi-channel array supplied - need binary array.", lvl=2)
        return

    if apply_drawing and "data" in annotation:

        drawings = annotation["data"][_vars._drawing_type]

        ## apply coords to tool and draw on canvas
        for coords in drawings:
            if len(coords) == 0:
                continue
            else:
                cv2.polylines(
                    image_bin, np.array([coords[0]]), False, coords[1], coords[2],
                )
                
                
    if not match_against.__class__.__name__ == "NoneType":
        match_image_bin = np.zeros(image_bin.shape, dtype="uint8")                        
        match_image_bin = visualization.draw_contour(
            image=match_image_bin, 
            annotations=annotations, 
            contour_id=match_against, 
            line_colour=255,
            line_width=0,
            fill=1)
        image_bin = cv2.bitwise_and(image_bin, match_image_bin)
        ul._print("- match detected contours against existing contour \"{}´\"".format(match_against))

    # =============================================================================
    # execute

    contours_det, hierarchies_det = cv2.findContours(
        image=image_bin,
        mode=_vars.opencv_contour_flags["retrieval"][retrieval],
        method=_vars.opencv_contour_flags["approximation"][approximation],
        offset=tuple(offset_coords),
    )

    # =============================================================================
    # process

    contours, support = [], []
    if len(contours_det) > 0:

        for idx, (contour, hierarchy) in enumerate(
            zip(contours_det, hierarchies_det[0])
        ):

            ## number of contour nodes
            if len(contour) > min_nodes and len(contour) < max_nodes:
                
                ## calc summary stats
                center, area, diameter = ul._calc_contour_stats(contour, stats_mode)
                
                ## contour hierarchy
                if hierarchy[3] == -1:
                    hierarchy_level = "parent"
                else:
                    hierarchy_level = "child"

                ## contour filter
                if all([
                        diameter > min_diameter and diameter < max_diameter,
                        area > min_area and area < max_area,
                    ]):

                    ## populate data lists
                    contours.append(contour)
                    support.append(
                        {
                            "center": center,
                            "area": area,
                            "diameter": diameter,
                            "hierarchy_level": hierarchy_level,
                            "hierarchy_idx_child": int(hierarchy[2]),
                            "hierarchy_idx_parent": int(hierarchy[3]),
                        }
                    )

        if len(contours) == 0:
            ul._print("- did not find any contours that match criteria", lvl=1)
        else:
            ul._print("- found " + str(len(contours)) + " contours that match criteria")
    else:
        ul._print("- did not find any contours", lvl=2)

    # =============================================================================
    # assemble results

    annotation = {
        "info": {
            "phenopype_function": fun_name,
            "phenopype_version": __version__,
            "annotation_type": annotation_type,
        },
        "settings": {
            "approximation": approximation,
            "retrieval": retrieval,
            "offset_coords": offset_coords,
            "min_nodes": min_nodes,
            "max_nodes": max_nodes,
            "min_area": min_area,
            "max_area": max_area,
            "min_diameter": min_diameter,
            "max_diameter": max_diameter,
        },
        "data": {
            "n": len(contours), 
            annotation_type: contours, 
            "support": support,},
    }

    # =============================================================================
    # return

    return ul._update_annotations(
        annotations=annotations,
        annotation=annotation,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )


def edit_contour(
    image,
    annotations,
    line_width="auto",
    overlay_blend=0.2,
    overlay_colour_left="default",
    overlay_colour_right="default",
    **kwargs,
):
    """
    Edit contours with a "paintbrush". The brush size can be controlled by pressing 
    Tab and using the mousewhell. Right-click removes, and left-click adds areas to 
    contours. Overlay colour, transparency (blend) and outline can be controlled.

    Parameters
    ----------
    image : ndarray
        input image
    annotations: dict
        phenopype annotation containing contours
    overlay_blend: float, optional
        transparency / colour-mixing of the contour overlay 
    line_width: int, optional
        add outline to the contours. useful when overlay_blend == 0
    left_colour: str, optional
        overlay colour for left click (include). (for options see pp.colour)
    right_colour: str, optional
        overlay colour for right click (exclude). (for options see pp.colour)
        
    Returns
    -------
    annotations: dict
        phenopype annotation containing contours

    """
    
    # =============================================================================
    # setup
    
    line_width = kwargs.get("overlay_line_width", line_width)
    
    
    # =============================================================================
    # annotation management

    ## get contours
    annotation_type = _vars._contour_type
    annotation_id = kwargs.get(annotation_type + "_id", None)
    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )
    gui_data = {annotation_type: ul._get_GUI_data(annotation)}

    ## get previous drawing
    fun_name = sys._getframe().f_code.co_name
    annotation_type = ul._get_annotation_type(fun_name)
    annotation_id = kwargs.get("annotation_id", None)
    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    gui_data.update({_vars._sequence_type: ul._get_GUI_data(annotation)})
    gui_settings = ul._get_GUI_settings(kwargs, annotation)


    # =============================================================================
    # execute

    gui = ul._GUI(
        image=image,
        tool="draw",
        line_width=line_width,
        overlay_blend=overlay_blend,
        overlay_colour_left=overlay_colour_left,
        overlay_colour_right=overlay_colour_right,
        data=gui_data,
        **gui_settings,
    )

    # =============================================================================
    # assemble results

    annotation = {
        "info": {
            "annotation_type": annotation_type,
            "phenopype_function": fun_name,
            "phenopype_version": __version__,
        },
        "settings": {
            "line_width": line_width,
            "overlay_blend": overlay_blend,
            "overlay_colour_left": overlay_colour_left,
            "overlay_colour_right": overlay_colour_right,
        },
        "data": {annotation_type: gui.data[_vars._sequence_type],},
    }

    # =============================================================================
    # return

    annotation = ul._update_annotations(
        annotations=annotations,
        annotation=annotation,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    if kwargs.get("ret_image", False):
        return annotation, gui.image_bin_copy
    else:
        return annotation


def mask_to_contour(
    annotations,
    include=True,
    box_margin=0,
    **kwargs,
):

    """
    Converts a mask to contour annotation, e.g. for the purpose of extracting information
    from this mask.  

    Parameters
    ----------
    annotation: dict
        phenopype annotation containing masks
    include : bool, optional
        include or exclude area inside contour
    box_margin : int, optional
        margin that is added between the outer perimeter of the mask and the box

    Returns
    -------
    annotation: dict
        phenopype annotation containing contours
    """
    ## fun name
    fun_name = sys._getframe().f_code.co_name

    # =============================================================================
    # annotation management

    ## get contours
    annotation_type = _vars._mask_type
    annotation_id = kwargs.get(annotation_type + "_id", None)
    
    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    masks = annotation["data"][annotation_type]
    
    annotation_type = ul._get_annotation_type(fun_name)
    annotation_id = kwargs.get("annotation_id", None)

    # =============================================================================
    # setup
        
    # =============================================================================
    # process

    contours, support = [], []
    for mask in masks:
        
        ## convert contour
        contour = ul._convert_tup_list_arr(mask)
        contours.append(contour)
        
        ## support variables
        center, radius = cv2.minEnclosingCircle(contour)
        center = [int(center[0]), int(center[1])]
        diameter = int(radius * 2)
        area = int(cv2.contourArea(contour))
        hierarchy_level = "parent"
        support.append(
            {
                "center": center,
                "area": area,
                "diameter": diameter,
                "hierarchy_level": hierarchy_level,
                "hierarchy_idx_child": "NA",
                "hierarchy_idx_parent": "NA",
            }
        )

    # =============================================================================
    # assemble results

    annotation = {
        "info": {
            "phenopype_function": fun_name,
            "phenopype_version": __version__,
            "annotation_type": annotation_type,
        },
        "settings": {
        },
        "data": {
            "n": len(masks),
            annotation_type: contours,
            "support": support,
            },
    }
    
    # =============================================================================
    # return
    
    return ul._update_annotations(
        annotations=annotations,
        annotation=annotation,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )


def morphology(
    image, kernel_size=5, shape="rect", operation="close", iterations=1, **kwargs
):
    """
    Performs advanced morphological transformations using erosion and dilation 
    as basic operations. Provides different kernel shapes and a suite of operation
    types (read more about morphological operations here:
    https://docs.opencv.org/master/db/df6/tutorial_erosion_dilatation.html)

    Parameters
    ----------
    image : ndarray
        input image (binary)
    kernel_size: int, optional
        size of the morphology kernel (has to be odd - even numbers will be ceiled)
    shape : {"rect", "cross", "ellipse"} str, optional
        shape of the kernel
    operation : {erode, dilate, open, close, gradient, tophat, blackhat, hitmiss} str, optional
        the morphology operation to be performed:
            - erode: remove pixels from the border 
            - dilate: add pixels to the border
            - open: erosion followed by dilation
            - close: dilation followed by erosion
            - gradient: difference between dilation and erosion of an input image
            - tophat: difference between input image and opening of input image
            - blackhat: difference between input image and closing of input image
            - hitmiss: find patterns in binary images (read more here:
              https://docs.opencv.org/master/db/d06/tutorial_hitOrMiss.html)
    iterations : int, optional
        number of times to run morphology operation

    Returns
    -------
    image : ndarray
        processed binary image

    """
    # =============================================================================
    # setup
    if kernel_size % 2 == 0:
        kernel_size = kernel_size + 1

    # =============================================================================
    # execute

    kernel = cv2.getStructuringElement(
        _vars.opencv_morphology_flags["shape_list"][shape],
        (kernel_size, kernel_size),
    )
    operation = _vars.opencv_morphology_flags["operation_list"][operation]

    image = cv2.morphologyEx(image, op=operation, kernel=kernel, iterations=iterations)

    # =============================================================================
    # return

    ## return
    return image


def threshold(
    image,
    method="otsu",
    constant=1,
    blocksize=99,
    value=127,
    mask=True,
    invert=False,
    **kwargs,
):
    """
    Applies a threshold to create a binary image from a grayscale or a multichannel 
    image (see phenopype.core.preprocessing.decompose_image for channel options).
    
    Three types of thresholding algorithms are supported: 
        - otsu: use Otsu algorithm to choose the optimal threshold value
        - adaptive: dynamic threshold values across image (uses arguments
          "blocksize" and "constant")
        - binary: fixed threshold value (uses argument "value")    
        
    Mask annotations can be supplied to include or exclude areas.

    Parameters
    ----------
    image : ndarray
        input image
    method : {"otsu", "adaptive", "binary"} str, optional
        type of thresholding algorithm to be used
    blocksize: int, optional
        Size of a pixel neighborhood that is used to calculate a threshold 
        value for the pixel (has to be odd - even numbers will be ceiled; for
        "adaptive" method)
    constant : int, optional
        value to subtract from binarization output (for "adaptive" method)
    value : {between 0 and 255} int, optional
        thesholding value (for "binary" method)
    channel {"gray", "red", "green", "blue"}: str, optional
        which channel of the image to use for thresholding 

    Returns
    -------
    image : ndarray
        binary image

    """

    # =============================================================================
    # setup

    ## set flags
    flags = make_dataclass(
        cls_name="flags",
        fields=[("mask", bool, mask), 
                ("invert", bool, invert),
                ],
    )

    if len(image.shape) == 3:
        if image.shape[2] > 1:
            ul._print("- multichannel image supplied - converting to single", lvl=1)
            image = preprocessing.decompose_image(image, col_space="gray")
        
    if blocksize % 2 == 0:
        blocksize = blocksize + 1
        ul._print("- even blocksize supplied, adding 1 to make odd", lvl=1)

    if flags.invert:
        image = cv2.bitwise_not(image)
        
    # =============================================================================
    # annotation management

    annotations = kwargs.get("annotations", {})

    ## references
    annotation_id_ref = kwargs.get(_vars._reference_type + "_id", None)
    annotation_ref = ul._get_annotation(
        annotations,
        _vars._reference_type,
        annotation_id_ref,
        prep_msg="- masking regions in thresholded image:",
    )

    ## masks
    annotation_id_mask = kwargs.get(_vars._mask_type + "_id", None)
    annotation_mask = ul._get_annotation(
        annotations,
        _vars._mask_type,
        annotation_id_mask,
        prep_msg="- masking regions in thresholded image:",
    )
           
    # =============================================================================
    # execute masking
    
    ## no masks / exclude
    height, width = image.shape
    rx, ry, rw, rh = 0, 0, width, height
    roi_list, roi_bbox_coords_list, roi_mask_coords_list = [image], [(rx, ry, rw, rh)], [""]
    
    ## with include masks 
    if all([flags.mask, "data" in annotation_mask]):
        if annotation_mask["data"]["include"]:
            roi_list, roi_bbox_coords_list, roi_mask_coords_list = [], [], []
            if len(annotation_mask["data"][_vars._mask_type]) > 0: 
                polygons = annotation_mask["data"][_vars._mask_type]   
                
                for coords in polygons:
                    
                    if type(coords) == list:
                        mask_coords = ul._convert_tup_list_arr(coords)
                    else:
                        mask_coords = coords
                        
                        
                    rx, ry, rw, rh = cv2.boundingRect(mask_coords)
                    
                    roi_list.append(image[ry : ry + rh, rx : rx + rw])
                    roi_bbox_coords_list.append((rx, ry, rw, rh))
                    roi_mask_coords_list.append(mask_coords)

    # =============================================================================
    # execute
    
    thresh = np.zeros(image.shape, dtype=np.uint8)
    
    for roi, roi_bbox_coords, roi_mask_coords in zip(
            roi_list, roi_bbox_coords_list, roi_mask_coords_list):
    
        if method == "otsu":
            ret, roi_thresh = cv2.threshold(
                roi, 
                0, 
                255, 
                cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
                )
        elif method == "adaptive":
            roi_thresh = cv2.adaptiveThreshold(
                roi,
                255,
                cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                cv2.THRESH_BINARY_INV,
                blocksize,
                constant,
                )
        elif method == "binary":
            ret, roi_thresh = cv2.threshold(
                roi, 
                value, 
                255,
                cv2.THRESH_BINARY_INV,
                )
        
        rx, ry, rw, rh = roi_bbox_coords
        
        if all([flags.mask, "data" in annotation_mask]):
            
            if all([annotation_mask["data"]["include"], 
                    annotation_mask["settings"]["tool"] == "polygon"]):
                
                poly_mask = np.zeros(roi.shape, dtype=np.uint8)
                poly_mask = cv2.drawContours(
                    image=poly_mask,
                    contours=[roi_mask_coords],
                    contourIdx=0,
                    thickness=-1,
                    color=255,
                    offset=(-rx, -ry),
                )
                roi_thresh = cv2.bitwise_and(roi_thresh, poly_mask)
        
        thresh[ry : ry + rh, rx : rx + rw] = roi_thresh
                
    # =============================================================================
    # exclude masks or references
    
    if "data" in annotation_ref:
        if _vars._mask_type in annotation_ref["data"]:
            polygons = annotation_ref["data"][_vars._mask_type]
            for coords in polygons:
                thresh[ul._create_mask_bool(thresh, coords)] = 0
            ul._print("- excluding pixels from reference")
                        
    ## with exclude masks 
    if all([flags.mask, "data" in annotation_mask]):
        if not annotation_mask["data"]["include"]:
            
            if len(annotation_mask["data"][_vars._mask_type]) > 0:
                polygons = annotation_mask["data"][_vars._mask_type]   
               
            for coords in polygons:
                thresh[ul._create_mask_bool(thresh, coords)] = 0
            ul._print("- excluding pixels from mask")

    # =============================================================================
    # return

    return thresh


def watershed(
    image,
    annotations,
    iterations=1,
    kernel_size=3,
    distance_cutoff=0.8,
    distance_mask=0,
    distance_type="l1",
    **kwargs,
):
    """
    Performs non-parametric marker-based segmentation - useful if many detected 
    contours are touching or overlapping with each other. Input image should be 
    a binary image that serves as the true background. Iteratively, edges are 
    eroded, the difference serves as markers.

    Parameters
    ----------
    image : ndarray
        input image
    image_thresh: array
        binary image (e.g. from threshold)
    kernel_size: int, optional
        size of the diff-kernel (has to be odd - even numbers will be ceiled)
    iterations : int, optional
        number of times to apply diff-operation
    distance_cutoff : {between 0 and 1} float, optional
        watershed distance transform cutoff (larger values discard more pixels)
    distance_mask : {0, 3, 5} int, optional
        size of distance mask - not all sizes work with all distance types (will
        be coerced to 0)
    distance_type : {"user", "l1", "l2", "C", "l12", "fair", "welsch", "huber"} str, optional
        distance transformation type

    Returns
    -------
    image : ndarray
        binary image

    """
    # =============================================================================
    # annotation management

    ## get contours
    annotation_type = _vars._contour_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    contours = annotation["data"][_vars._contour_type]

    # =============================================================================
    # setup

    if len(contours) > 0:

        ## coerce to multi channel image for colour mask
        if len(image.shape) == 2:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

        ## create binary overlay
        image_bin = np.zeros(image.shape[0:2], dtype=np.uint8)

        ## draw contours onto overlay
        for contour in contours:
            cv2.drawContours(
                image=image_bin,
                contours=[contour],
                contourIdx=0,
                thickness=-1,
                color=255,
                maxLevel=3,
                offset=(0, 0),
            )

    if kernel_size % 2 == 0:
        kernel_size = kernel_size + 1

    # =============================================================================
    # execute

    ## sure background
    sure_bg = morphology(
        image_bin,
        operation="dilate",
        shape="ellipse",
        kernel_size=kernel_size,
        iterations=iterations,
    )

    ## distances in foreground
    if distance_type in ["user", "l12", "fair", "welsch", "huber"]:
        distance_mask = 0

    opened = morphology(
        image_bin,
        operation="erode",
        shape="ellipse",
        kernel_size=kernel_size,
        iterations=iterations,
    )

    ## distance transformation
    dist_transform = cv2.distanceTransform(
        opened, _vars.opencv_distance_flags[distance_type], distance_mask
    )
    dist_transform = cv2.normalize(
        dist_transform, dist_transform, 0, 1.0, cv2.NORM_MINMAX
    )
    dist_transform = preprocessing.blur(
        dist_transform, kernel_size=int(2 * kernel_size)
    )
    dist_transform = abs(255 * (1 - dist_transform))
    dist_transform = dist_transform.astype(np.uint8)

    ## sure foreground
    sure_fg = threshold(
        dist_transform, method="binary", value=int(distance_cutoff * 255)
    )

    ## finding unknown region
    sure_fg = sure_fg.astype("uint8")
    sure_fg[sure_fg == 1] = 255
    unknown = cv2.subtract(sure_bg, sure_fg)

    ## marker labelling
    ret, markers = cv2.connectedComponents(sure_fg)
    markers = markers + 1
    markers[unknown == 255] = 0

    ## watershed
    markers = cv2.watershed(
        preprocessing.blur(image, int(2 * kernel_size + 1)), markers
    )

    ## convert to contours
    watershed_mask = np.zeros(image.shape[:2], np.uint8)
    watershed_mask[markers == -1] = 255
    watershed_mask[0 : watershed_mask.shape[0], 0] = 0
    watershed_mask[0 : watershed_mask.shape[0], watershed_mask.shape[1] - 1] = 0
    watershed_mask[0, 0 : watershed_mask.shape[1]] = 0
    watershed_mask[watershed_mask.shape[0] - 1, 0 : watershed_mask.shape[1]] = 0

    annotations_contour = detect_contour(watershed_mask, retrieval="ccomp")
    image_watershed = np.zeros(watershed_mask.shape, np.uint8)

    for coord, supp in zip(
        annotations_contour[annotation_type]["a"]["data"][annotation_type],
        annotations_contour[annotation_type]["a"]["data"]["support"],
    ):
        if supp["hierarchy_level"] == "child":
            cv2.drawContours(
                image=image_watershed,
                contours=[coord],
                contourIdx=0,
                thickness=-1,
                color=ul._get_bgr("white"),
                maxLevel=3,
                offset=None,
            )
            cv2.drawContours(
                image=image_watershed,
                contours=[coord],
                contourIdx=0,
                thickness=2,
                color=ul._get_bgr("black"),
                maxLevel=3,
                offset=None,
            )

    # =============================================================================
    # return

    return image_watershed


# def watershed(
#     image,
#     annotations,
#     iterations=1,
#     kernel_size=3,
#     distance_cutoff=0.8,
#     distance_mask=0,
#     distance_type="l1",
#     **kwargs
# ):
#     """
#     Performs non-parametric marker-based segmentation - useful if many detected
#     contours are touching or overlapping with each other. Input image should be
#     a binary image that serves as the true background. Iteratively, edges are
#     eroded, the difference serves as markers.

#     Parameters
#     ----------
#     image : ndarray
#         input image
#     image_thresh: array
#         binary image (e.g. from threshold)
#     kernel_size: int, optional
#         size of the diff-kernel (has to be odd - even numbers will be ceiled)
#     iterations : int, optional
#         number of times to apply diff-operation
#     distance_cutoff : {between 0 and 1} float, optional
#         watershed distance transform cutoff (larger values discard more pixels)
#     distance_mask : {0, 3, 5} int, optional
#         size of distance mask - not all sizes work with all distance types (will
#         be coerced to 0)
#     distance_type : {"user", "l1", "l2", "C", "l12", "fair", "welsch", "huber"} str, optional
#         distance transformation type

#     Returns
#     -------
#     image : ndarray
#         binary image

#     """
#     # =============================================================================
#     # annotation management

#     ## get contours
#     annotation_type = _vars._contour_type
#     annotation_id = kwargs.get(annotation_type + "_id", None)

#     annotation = ul._get_annotation(
#         annotations=annotations,
#         annotation_type=annotation_type,
#         annotation_id=annotation_id,
#         kwargs=kwargs,
#     )

#     contours = annotation["data"][_vars._contour_type]

# 	# =============================================================================
# 	# setup

#     if len(contours) > 0:

#         ## coerce to multi channel image for colour mask
#         if len(image.shape) == 2:
#             image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

#         ## create binary overlay
#         image_bin = np.zeros(image.shape[0:2], dtype=np.uint8)

#         ## draw contours onto overlay
#         for contour in contours:
#             cv2.drawContours(
#                 image=image_bin,
#                 contours=[contour],
#                 contourIdx=0,
#                 thickness=-1,
#                 color=255,
#                 maxLevel=3,
#                 offset=(0,0),
#                 )


#     if kernel_size % 2 == 0:
#         kernel_size = kernel_size + 1

# 	# =============================================================================
# 	# execute

#     ## sure background
#     sure_bg = morphology(
#         image_bin,
#         operation="dilate",
#         shape="ellipse",
#         kernel_size=kernel_size,
#         iterations=iterations,
#     )

#     ## distances in foreground
#     if distance_type in ["user", "l12", "fair", "welsch", "huber"]:
#         distance_mask = 0

#     opened = morphology(
#         image_bin,
#         operation="erode",
#         shape="ellipse",
#         kernel_size=kernel_size,
#         iterations=iterations,
#     )

#     ## distance transformation
#     dist_transform = cv2.distanceTransform(
#         opened, _vars.opencv_distance_flags[distance_type], distance_mask
#     )
#     dist_transform = cv2.normalize(
#         dist_transform, dist_transform, 0, 1.0, cv2.NORM_MINMAX
#     )
#     dist_transform = preprocessing.blur(dist_transform, kernel_size=int(2*kernel_size))
#     dist_transform = abs(255 * (1-dist_transform))
#     dist_transform = dist_transform.astype(np.uint8)

#     ## sure foreground
#     sure_fg = threshold(dist_transform, method="binary", value=int(distance_cutoff*255))

#     ## finding unknown region
#     sure_fg = sure_fg.astype("uint8")
#     sure_fg[sure_fg == 1] = 255
#     unknown = cv2.subtract(sure_bg, sure_fg)

#     ## marker labelling
#     ret, markers = cv2.connectedComponents(sure_fg)
#     markers = markers + 1
#     markers[unknown == 255] = 0

#     ## watershed
#     markers = cv2.watershed(preprocessing.blur(image, int(2*kernel_size+1)), markers)

#     ## convert to contours
#     watershed_mask = np.zeros(image.shape[:2], np.uint8)
#     watershed_mask[markers == -1] = 255
#     watershed_mask[0 : watershed_mask.shape[0], 0] = 0
#     watershed_mask[0 : watershed_mask.shape[0], watershed_mask.shape[1] - 1] = 0
#     watershed_mask[0, 0 : watershed_mask.shape[1]] = 0
#     watershed_mask[watershed_mask.shape[0] - 1, 0 : watershed_mask.shape[1]] = 0

#     annotations_contour = detect_contour(watershed_mask, retrieval="ccomp")
#     image_watershed = np.zeros(watershed_mask.shape, np.uint8)

#     for coord, supp in zip(annotations_contour[annotation_type]["a"]["data"][annotation_type],
#                            annotations_contour[annotation_type]["a"]["data"]["support"]):
#         if supp["hierarchy_level"] == "child":
#             cv2.drawContours(
#                 image=image_watershed,
#                 contours=[coord],
#                 contourIdx=0,
#                 thickness=-1,
#                 color=ul._get_bgr("white"),
#                 maxLevel=3,
#                 offset=None
#                 )
#             cv2.drawContours(
#                 image=image_watershed,
#                 contours=[coord],
#                 contourIdx=0,
#                 thickness=2,
#                 color=ul._get_bgr("black"),
#                 maxLevel=3,
#                 offset=None
#                 )

# 	# =============================================================================
# 	# return

#     return image_watershed



# -*- coding: utf-8 -*-
"""
Created on Tue May 14 20:02:39 2024

@author: mluerig
"""
import io
import os
import warnings
from contextlib import redirect_stdout
from functools import wraps

from phenopype import utils_lowlevel as ul

#%% decorators 

def annotation_function(fun):
    
    @wraps(fun)
    def annotation_function_wrapper(*args, **kwargs):

        ## determine the annotation type from function name
        kwargs["annotation_type"] = ul._get_annotation_type(fun.__name__)

        ## get annotation using 
        if "annotations" in kwargs:
            if isinstance(kwargs["annotations"], dict):
                if len(kwargs["annotations"]) > 0:
                    kwargs["annotation_id"] = ul._get_annotation_id(**kwargs)
                    kwargs["annotation"] = ul._get_annotation2(**kwargs)
                
        ## run function
        kwargs["annotation"] = fun(*args, **kwargs)
    
        ## return and update annotations    
        if "annotations" in kwargs:
            result = ul._update_annotations(**kwargs)
        else:
            result = kwargs["annotation"]

        return result
    
    ## close function wrapper
    return annotation_function_wrapper

# def capture_stdout_log(fun, logger, *logger_args):
    
#     @wraps(fun)
#     def capture_stdout_log_wrapper():
                
#         string_buffer = io.StringIO()
#         with redirect_stdout(string_buffer):  
#             fun()
    
#         ## reformat stdout
#         stdout = string_buffer.getvalue()
#         stdout_list = stdout.split("\n")
#         for line in stdout_list:
#             if not line == "":
#                 if line.endswith("\n"):
#                     line = line[:-2]
#                 logger(logger_args)
                
#     ## close function wrapper
#     return capture_stdout_log_wrapper

# def capture_print(func):
#     def wrapper(*args, **kwargs):
#         # Redirect stdout to capture print output
#         old_stdout = sys.stdout
#         sys.stdout = captured_output = StringIO()

#         try:
#             result = func(*args, **kwargs)
#             output = captured_output.getvalue()
#         finally:
#             # Reset stdout
#             sys.stdout = old_stdout
        
#         return result, output

#     return wrapper

def deprecation_warning(new_func=None):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn(f"{func.__name__} is deprecated; use {new_func.__name__} instead.", category=DeprecationWarning, stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)  # reset filter
            return new_func(*args, **kwargs)
        return wrapper
    return decorator
    

def legacy_args(fun):
    @wraps(fun)
    def wrapper(*args, **kwargs):
        if fun.__name__ == "save_image":
            if 'file_path' not in kwargs or not kwargs['file_path']:
                file_name = kwargs.get('file_name')
                dir_path = kwargs.get('dir_path')
                if file_name and dir_path:
                    kwargs['file_path'] = os.path.join(dir_path, file_name)
                    ul._print(f"WARNING - using 'file_name' and 'dir_path' is deprecated for >{fun.__name__}< - use 'file_path' instead", lvl=1)
        
        elif fun.__name__ == "save_canvas":
            if 'file_path' not in kwargs or not kwargs['file_path']:
                file_name = kwargs.get('file_name')
                dir_path = kwargs.get('dir_path')
                if file_name and dir_path:
                    kwargs['file_path'] = os.path.join(dir_path, file_name)
                    print(f"Warning - using file_name and dir_path is deprecated for {fun.__name__} - use file_path instead")
        
        return fun(*args, **kwargs)
    
    return wrapper


from collections import defaultdict

gui_zoom_config = None
last_print_msg = ""
pype_restart = False

models = defaultdict(dict)
reference_templates = defaultdict(dict)

template_loaded_current = None
template_path_current = None

verbose = False
verbosity_level = 0  # 0 = all, 1 = warnings, 2 = errors (non-breaking), 3 = errors (breaking)

min_visible_px = 1
max_linewidth_px = 10
instructions_show = True
instructions_pos = (0.5,0.9)
window_close = False
window_max_dim = 1000
window_min_dim = 100




import importlib.metadata
import warnings

__version__ = importlib.metadata.version("phenopype")

class _PluginsPlaceholder:
    def __init__(self, message="The plugins module could not be imported - is it installed?"):
        self.message = message

    def __getattr__(self, name):
        warnings.warn(self.message, UserWarning)

    def __call__(self, *args, **kwargs):
        warnings.warn(self.message, UserWarning)

from . import utils_lowlevel
from . import config, decorators
from .core import preprocessing, segmentation, measurement, export, visualization
from .main import Project, Project_labelling, Pype
from .utils import load_image, show_image, print_colours, save_image, resize_image
from .tracking import motion_tracker, tracking_method




#%% modules

import cv2
import io
import os
import sys
import webbrowser

from dataclasses import make_dataclass
from contextlib import redirect_stdout
from pkg_resources import resource_filename
import ruamel.yaml

from phenopype import _vars
from phenopype import config
from phenopype import decorators
from phenopype import utils_lowlevel as ul


#%% functions


def load_image(path, mode="unchanged", **kwargs):
    """
    Create ndarray from image path or return or resize exising array.

    Parameters
    ----------
    path: str
        path to an image stored on the harddrive
    mode: {"default", "colour","gray"} str, optional
        image conversion on loading:
            - default: load image as is
            - colour: convert image to 3-channel (BGR)
            - gray: convert image to single channel (grayscale)
    kwargs:
        developer options

    Returns
    -------
    container: container
        A phenopype container is a Python class where loaded images,
        dataframes, detected contours, intermediate output, etc. are stored
        so that they are available for inspection or storage at the end of
        the analysis.
    image: ndarray
        original image (resized, if selected)

    """

    ## set flags
    flags = make_dataclass(cls_name="flags", fields=[("mode", str, mode)])

    ## load image
    if path.__class__.__name__ == "str":
        if os.path.isfile(path):
            ext = os.path.splitext(path)[1]
            if ext.replace(".", "") in _vars.default_filetypes:
                if flags.mode == "unchanged":
                    image = cv2.imread(path, cv2.IMREAD_UNCHANGED)
                elif flags.mode == "colour":
                    image = cv2.imread(path, cv2.IMREAD_COLOR)
                elif flags.mode == "gray":
                    image = cv2.imread(path, cv2.IMREAD_GRAYSCALE)
                elif flags.mode == "rgb":
                    image = cv2.imread(path, cv2.IMREAD_UNCHANGED)
                    if image.shape[2] == 4:
                        b, g, r, a = cv2.split(image)
                        image = cv2.merge((r, g, b, a))
                    else:
                        b, g, r = cv2.split(image)
                        image = cv2.merge((r, g, b))
            else:
                ul._print(
                    'Invalid file extension "{}" - could not load image:\n'.format(ext)
                )
                return
        elif os.path.isdir(path):
            image = ul._load_project_image_directory(
                path, as_container=False
            )
        else:
            ul._print("Invalid image path - could not load image.")
            return
    else:
        ul._print("Invalid input format - could not load image.")
        return

    return image


def print_colours():

    colours_path = os.path.join(resource_filename("phenopype", "assets"), "wc3_colours.html")
    webbrowser.open_new_tab(colours_path)


def resize_image(
        image, 
        factor=1, 
        factor_ret=False,
        width=None,
        height=None,
        max_dim=None, 
        interpolation="cubic",
        ):
    """
    Resize image by resize factor 

    Parameters
    ----------
    image: array 
        Image to be resized
    factor: float, optional
        Resize factor for the image (1 = 100%, 0.5 = 50%, 0.1 = 10% of original size). At 1, no resizing will be performed
    factor_ret: bool, optional
        If True, returns the image and the resize factor
    width: int, optional
        Width to resize the image to. If specified, height must also be specified
    height: int, optional
        Height to resize the image to. If specified, width must also be specified
    max_dim: int, optional
        Maximum size of any dimension that the image will be resized to. Maintains aspect ratio
    interpolation: {'nearest', 'linear', 'cubic', 'area', 'lanczos', 'lin_exact', 'inter', 'warp_fill', 'warp_inverse'} str, optional
        interpolation algorithm to use - refer to https://docs.opencv.org/4.9.0/da/d54/group__imgproc__transform.html#ga5bb5a1fea74ea38e1a5445ca803ff121

    Returns
    -------
    image : ndarray
        resized image

    """
    
    image_height, image_width = image.shape[:2]

    if width is not None and height is not None:
        image = cv2.resize(
            image,
            (width, height),
            interpolation=_vars.opencv_interpolation_flags[interpolation]
        )
    elif max_dim is not None:
        if image_height > max_dim or image_width > max_dim:
            if image_width >= image_height:
                factor = max_dim / image_width
            else:
                factor = max_dim / image_height
            new_image_width = int(image_width * factor)
            new_image_height = int(image_height * factor)
            image = cv2.resize(
                image,
                (new_image_width, new_image_height),
                interpolation=_vars.opencv_interpolation_flags[interpolation]
            )
    elif factor != 1:
        image = cv2.resize(
            image,
            (0, 0),
            fx=factor,
            fy=factor,
            interpolation=_vars.opencv_interpolation_flags[interpolation]
        )

    if factor_ret:
        return image, factor
    return image

@decorators.legacy_args
def save_image(image, file_path, suffix=None, overwrite=False, **kwargs):
    """Save an image (array) to a specified format.

    Parameters
    ----------
    image : array
        Image to save.
    file_path : str
        Full path (including directory, base name, and extension) to save the image.
    suffix : str, optional
        Suffix to append to the image name to prevent overwriting.
    overwrite : bool, optional
        If True, overwrite existing files with the same name.

    Returns
    -------
    bool
        True if the image was saved successfully, False otherwise.
    """
    # Normalize file path and extract base name and extension
    dir_path, file_name = os.path.split(file_path)
    base_name, ext = os.path.splitext(file_name)
    
    # Handle suffix
    suffix = f"_{suffix}" if suffix else ""
    
    # Construct new file path
    file_name_new = f"{base_name}{suffix}{ext}"
    new_file_path = os.path.join(dir_path, file_name_new)

    # Check if file exists and handle overwrite logic
    if ul._overwrite_check(new_file_path, overwrite):
        success = cv2.imwrite(new_file_path, image)
        return success



def show_image(
    image,
    return_input=False,
    position_reset=True,
    position_offset=25,
    window_aspect="normal",
    check=True,
    **kwargs
):
    """
    Show one or multiple images by providing path string or array or list of
    either.

    Parameters
    ----------
    image: array, list of arrays
        the image or list of images to be displayed. can be array-type,
        or list or arrays
    window_max_dim: int, optional
        maximum dimension on either acis
    window_aspect: {"fixed", "free"} str, optional
        type of opencv window ("free" is resizeable)
    position_reset: bool, optional
        flag whether image positions should be reset when reopening list of
        images
    position_offset: int, optional
        if image is list, the distance in pixels betweeen the positions of
        each newly opened window (only works in conjunction with
        "position_reset")
    check: bool, optional
        user input required when more than 10 images are opened at the same
        time
    """
    ## kwargs
    flag_check = check

    ## load image
    if image.__class__.__name__ == "ndarray":
        pass
    elif image.__class__.__name__ == "list":
        pass
    else:
        ul._print("wrong input format.")
        return

    ## open images list or single images
    while True:
        if isinstance(image, list):
            if len(image) > 10 and flag_check == True:
                warning_string = (
                    "WARNING: trying to open "
                    + str(len(image))
                    + " images - proceed (y/n)?"
                )
                check = input(warning_string)
                if check in ["y", "Y", "yes", "Yes"]:
                    ul._print("Proceed - Opening images ...")
                    pass
                else:
                    ul._print("Aborting")
                    break
            idx = 0
            for i in image:
                idx += 1
                if i.__class__.__name__ == "ndarray":
                    ul._print("phenopype" + " - " + str(idx))
                    ul._GUI(
                        i,
                        mode="",
                        window_aspect=window_aspect,
                        window_name="phenopype" + " - " + str(idx),
                        window_control="external",
                        **kwargs,
                    )
                    if position_reset == True:
                        cv2.moveWindow(
                            "phenopype" + " - " + str(idx),
                            int(idx + idx * position_offset),
                            int(idx + idx * position_offset),
                        )
                else:
                    ul._print("skipped showing list item of type " + i.__class__.__name__)
            cv2.waitKey(0)
            cv2.destroyAllWindows()
            break
        else:
            out = ul._GUI(
                image=image,
                mode="",
                return_input=return_input,
                window_aspect=window_aspect,
                window_name="phenopype",
                window_control="internal",
                **kwargs,
            )
            if return_input:
                return out.keypress_trans
            break



#%% modules

import mock
import pytest

import phenopype as pp


#%% fixtures
@pytest.fixture
def canvas(image):
    
    canvas = pp.visualization.select_canvas(
        image, 
        )
    
    return canvas

@pytest.fixture
def container():
    with mock.patch('builtins.input', return_value='y'):
        project = pp.Project(root_dir=pytest.project_root_dir_1)    
        
    project.add_files(
        image_dir=pytest.image_dir, 
        mode="link", 
        include="stickle"
        )
        
    return pp.utils_lowlevel._load_project_image_directory(project.dir_paths[0])

#%% tests

def test_select_canvas(image, image_binary, container):
    
    canvas = pp.visualization.select_canvas(
        image_binary, 
        canvas="raw", 
        multi=True,
        )
    
    canvas = pp.visualization.select_canvas(
        container, 
        canvas="mod", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        container, 
        canvas="gray", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        container, 
        canvas="red", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        container, 
        canvas="green", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        container, 
        canvas="blue", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        container, 
        canvas="raw", 
        multi=True,
        )
    
    
    canvas = pp.visualization.select_canvas(
        image, 
        canvas="mod", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        image, 
        canvas="gray", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        image, 
        canvas="red", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        image, 
        canvas="green", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        image, 
        canvas="blue", 
        multi=True,
        )
    canvas = pp.visualization.select_canvas(
        image, 
        canvas="raw", 
        multi=True,
        )
    
    assert (image==canvas).all()



def test_draw_contour(canvas, contours):
    
    annotations = contours
    
    canvas_mod = pp.visualization.draw_contour(
        canvas,
        annotations,
        bounding_box=True,
        label=True,
        )
    
    assert not (canvas==canvas_mod).all()



def test_draw_landmarks(canvas, landmarks):
    
    annotations = landmarks
    
    canvas_mod = pp.visualization.draw_landmark(
        canvas,
        annotations,
        )
    
    assert not (canvas==canvas_mod).all()
    
    

def test_draw_mask(canvas, mask_polygon):
    
    annotations = mask_polygon
    
    canvas_mod = pp.visualization.draw_mask(
        canvas,
        annotations,
        label=True,
        )
    
    assert not (canvas==canvas_mod).all()
    
    
    
def test_draw_polyline(canvas, polyline):
    
    annotations = polyline
    
    canvas_mod = pp.visualization.draw_polyline(
        canvas,
        annotations,
        )
    
    assert not (canvas==canvas_mod).all()



def test_draw_reference(canvas, reference_created, reference_detected):
    
    annotations = reference_created 
    
    canvas_mod = pp.visualization.draw_reference(
        canvas,
        annotations,
        label=True,
        )
    
    annotations = reference_detected 
        
    canvas_mod = pp.visualization.draw_reference(
        canvas,
        annotations,
        label=True,
        )
    
    assert not (canvas==canvas_mod).all()






#%% modules

import copy
import cv2
import numpy as np
import math
import sys
from dataclasses import make_dataclass

from phenopype import _vars, config
from phenopype import utils_lowlevel as ul


#%% settings

inf = math.inf

#%% functions

def draw_comment(
    image,
    annotations,
    label_colour="default",
    label_size="auto",
    label_width="auto",
    background=True,
    background_colour="white",
    background_pad=10,
    background_border="black",
    font="simplex",
    **kwargs,
):   
    """

    Parameters
    ----------
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing QR-code (comment)
    line_colour: {"default", ... see phenopype.print_colours()} str, optional
        contour line colour - default colour as specified in settings
    line_width: {"auto", ... int > 0} int, optional 
        contour line width - automatically scaled to image by default
    label : bool, optional
        draw reference label
    label_colour : {"default", ... see phenopype.print_colours()} str, optional
        contour label colour - default colour as specified in settings
    label_size: {"auto", ... int > 0} int, optional 
        contour label font size - automatically scaled to image by default
    label_width:  {"auto", ... int > 0} int, optional 
        contour label font thickness - automatically scaled to image by default
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    image: ndarray
        canvas with contours

    """
    # =============================================================================
    # setup
    
        
    label_size = ul._get_size(image.shape[1], image.shape[0], "label_size", label_size)
    label_width = ul._get_size(image.shape[1], image.shape[0], "label_width", label_width)
    
    label_colour = ul._get_bgr(label_colour, "label_colour")

    font = _vars.opencv_font_flags[font]

    if background:
        background_colour = ul._get_bgr(background_colour)
        background_border = ul._get_bgr(background_border)

    # =============================================================================
    # annotation management

    annotation_type = _vars._comment_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    # =============================================================================
    # execute
    
    canvas = copy.deepcopy(image)
    
    ## prep label
    name = annotation["data"]["label"]
    label_text = annotation["data"][_vars._comment_type]
    label =  name + ": " + label_text
    label_coords = (int(canvas.shape[0] // 10), int(canvas.shape[1] / 3))
            
    ## add background
    if background:
        
        text_size, _ = cv2.getTextSize(label, font, label_size, label_width)
        text_size = int(text_size[0] + background_pad), int(text_size[1] + (text_size[1]/8) + background_pad)
                
        background_coords = label_coords
        background_coords = int(background_coords[0] - (background_pad)), \
            int(background_coords[1] + (text_size[1]/8) + (background_pad))

        background_border_line_width = ul._get_size(image.shape[1], image.shape[0], "line_width")

        cv2.rectangle(
            canvas, 
            background_coords, 
            (label_coords[0] + text_size[0], label_coords[1] - text_size[1]), 
            background_colour, 
            -1
        )
        
        cv2.rectangle(
            canvas, 
            background_coords, 
            (label_coords[0] + text_size[0], label_coords[1] - text_size[1]), 
            background_border, 
            background_border_line_width
        )
        
    ## draw label
    cv2.putText(
        canvas,
        label,
        label_coords,
        font,
        label_size,
        label_colour,
        label_width,
        cv2.LINE_AA,
    )

    # =============================================================================
    # return

    return canvas

def draw_contour(
    image,
    annotations,
    fill=0.3,
    line_colour="default",
    line_width="auto",
    label=False,
    label_colour="default",
    label_size="auto",
    label_width="auto",
    offset_coords=None,
    bbox=False,
    bbox_ext=20,
    bbox_colour="default",
    bbox_line_width="auto",
    **kwargs,
):
    """
    Draw contours and their labels onto a canvas. Can be filled or empty, offset
    coordinates can be supplied. 

    Parameters
    ----------
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing contours
    offset_coords : tuple, optional
        offset coordinates, will be added to all contours
    label : bool, optional
        draw contour label
    fill : float, optional
        background transparency for contour fill (0=no fill).
    level : int, optional
        the default is 3.
    line_colour: {"default", ... see phenopype.print_colours()} str, optional
        contour line colour - default colour as specified in settings
    line_width: {"auto", ... int > 0} int, optional 
        contour line width - automatically scaled to image by default
    label_colour : {"default", ... see phenopype.print_colours()} str, optional
        contour label colour - default colour as specified in settings
    label_size: {"auto", ... int > 0} int, optional 
        contour label font size - automatically scaled to image by default
    label_width:  {"auto", ... int > 0} int, optional 
        contour label font thickness - automatically scaled to image by default
    bounding_box: bool, optional
        draw bounding box around the contour
    bounding_box_ext: in, optional
        value in pixels by which the bounding box should be extended
    bounding_box_colour: {"green", "red", "blue", "black", "white"} str, optional
        bounding box line colour
    bounding_box_line_width: int, optional
        bounding box line width
        
    Returns
    -------
    image: ndarray
        canvas with contours

    """

    # =============================================================================
    # annotation management

    annotation_type = _vars._contour_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    contours = annotation["data"][annotation_type]
    contours_support = annotation["data"]["support"]
    
    if "contour_idx" in kwargs:
        contour_idx = kwargs.get("contour_idx")
        if contour_idx.__class__.__name__ == "int":
            contour_idx = [contour_idx]
        contours = [contours[i-1] for i in contour_idx]
        contours_support = [contours_support[i-1] for i in contour_idx]

    # =============================================================================
    # setup

    level = kwargs.get("level", 3)
    fill_colour = kwargs.get("fill_colour", line_colour)

    ## flags
    flags = make_dataclass(
        cls_name="flags",
        fields=[
            ("bbox", bool, bbox),
            ("label", bool, label),
            ("fill", bool, True),
        ],
    )

    line_width = ul._get_size(image.shape[1], image.shape[0], "line_width", line_width)
    label_size = ul._get_size(image.shape[1], image.shape[0], "label_size", label_size)
    label_width = ul._get_size(image.shape[1], image.shape[0], "label_width", label_width)
    bbox_line_width = ul._get_size(image.shape[1], image.shape[0], "line_width", bbox_line_width)

    fill_colour = ul._get_bgr(fill_colour, "line_colour")
    line_colour = ul._get_bgr(line_colour, "line_colour")
    label_colour = ul._get_bgr(label_colour, "label_colour")
    bbox_colour = ul._get_bgr(bbox_colour, "line_colour")

    ## filling and line settings
    if fill > 0:
        flags.fill = True
    else:
        flags.fill = False

    # =============================================================================
    # execute

    canvas = copy.deepcopy(image)

    ## 1) fill contours
    if flags.fill:
        colour_mask = copy.deepcopy(canvas)
        for contour in contours:
            cv2.drawContours(
                image=canvas,
                contours=[contour],
                contourIdx=0,
                thickness=-1,
                color=line_colour,
                maxLevel=level,
                offset=offset_coords,
            )
        canvas = cv2.addWeighted(colour_mask, 1 - fill, canvas, fill, 0)

    ## 2) contour lines
    for contour in contours:
        cv2.drawContours(
            image=canvas,
            contours=[contour],
            contourIdx=0,
            thickness=line_width,
            color=line_colour,
            maxLevel=level,
            offset=offset_coords,
        )

    ## 3) bounding boxes
    if flags.bbox:
        q = bbox_ext
        for contour in contours:
            rx, ry, rw, rh = cv2.boundingRect(contour)
            cv2.rectangle(
                canvas,
                (rx - q, ry - q),
                (rx + rw + q, ry + rh + q),
                bbox_colour,
                bbox_line_width,
            )

    ## 4) contour label
    if flags.label:
        for idx, support in enumerate(contours_support):
            cv2.putText(
                canvas,
                str(idx + 1),
                tuple(support["center"]),
                cv2.FONT_HERSHEY_SIMPLEX,
                label_size,
                label_colour,
                label_width,
                cv2.LINE_AA,
            )

    # =============================================================================
    # return

    return canvas


def draw_landmark(
    image,
    annotations,
    label=True,
    label_colour="default",
    label_size="auto",
    label_width="auto",
    offset=0,
    point_colour="default",
    point_size="auto",
    **kwargs,
):
    """
    Draw landmarks into an image.

    Parameters
    ----------
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing landmarks
    label : bool, optional
        draw landmark label
    label_colour : {"default", ... see phenopype.print_colours()} str, optional
        contour label colour - default colour as specified in settings
    label_size: {"auto", ... int > 0} int, optional 
        contour label font size - automatically scaled to image by default
    label_width:  {"auto", ... int > 0} int, optional 
        contour label font thickness - automatically scaled to image by default
    offset: int, optional
        add offset (in pixels) to text location (to bottom-left corner of the text string)
    point_colour: {"green", "red", "blue", "black", "white"} str, optional
        landmark point colour
    point_size: int, optional
        landmark point size in pixels

    Returns
    -------
    image: ndarray
        canvas with landmarks

    """

    # =============================================================================
    # annotation management

    annotation_type = _vars._landmark_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    points = annotation["data"][annotation_type]

    # =============================================================================
    # setup

    ## flags
    flags = make_dataclass(cls_name="flags", fields=[("label", bool, label)])
    
    point_size = ul._get_size(image.shape[1], image.shape[0], "point_size", point_size)
    label_size = ul._get_size(image.shape[1], image.shape[0], "label_size", label_size)
    label_width = ul._get_size(image.shape[1], image.shape[0], "label_width", label_width)

    point_colour = ul._get_bgr(point_colour, "point_colour")
    label_colour = ul._get_bgr(label_colour, "label_colour")
    
    # =============================================================================
    # execute

    canvas = copy.deepcopy(image)

    for idx, point in enumerate(points):
        x, y = point
        cv2.circle(canvas, (x, y), point_size, point_colour, -1)
        if flags.label:
            x, y = x + offset, y + offset
            cv2.putText(
                canvas,
                str(idx + 1),
                (x, y),
                cv2.FONT_HERSHEY_SIMPLEX,
                label_size,
                label_colour,
                label_width,
                cv2.LINE_AA,
            )

    # =============================================================================
    # return

    return canvas


def draw_mask(
    image,
    annotations,
    line_colour="default",
    line_width="auto",
    fill=0,
    label=False,
    label_colour="default",
    label_size="auto",
    label_width="auto",
    **kwargs,
):
    """
    Draw masks into an image. This function is also used to draw the perimeter 
    of a created or detected reference card.
    
    Parameters
    ----------        
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing masks
    label : bool, optional
        draw mask label
    line_colour: {"default", ... see phenopype.print_colours()} str, optional
        contour line colour - default colour as specified in settings
    line_width: {"auto", ... int > 0} int, optional 
        contour line width - automatically scaled to image by default
    label_colour : {"default", ... see phenopype.print_colours()} str, optional
        contour label colour - default colour as specified in settings
    label_size: {"auto", ... int > 0} int, optional 
        contour label font size - automatically scaled to image by default
    label_width:  {"auto", ... int > 0} int, optional 
        contour label font thickness - automatically scaled to image by default

    Returns
    -------
    image: ndarray
        canvas with masks
    """

    # =============================================================================
    # setup

    ## flags
    flags = make_dataclass(cls_name="flags", fields=[("label", bool, label)])

    # =============================================================================
    # annotation management

    annotation_type = _vars._mask_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    polygons = annotation["data"][annotation_type]
    label = annotation["data"]["label"]

    # =============================================================================
    # setup

    line_width = ul._get_size(image.shape[1], image.shape[0], "line_width", line_width)
    label_size = ul._get_size(image.shape[1], image.shape[0], "label_size", label_size)
    label_width = ul._get_size(image.shape[1], image.shape[0], "label_width", label_width)
    
    line_colour = ul._get_bgr(line_colour, "line_colour")
    label_colour = ul._get_bgr(label_colour, "label_colour")

    # =============================================================================
    # execute

    canvas = copy.deepcopy(image)
    
    for coords in polygons:
        if fill > 0:
            cv2.fillPoly(canvas, np.array([coords]), line_colour)
        else:
            cv2.polylines(canvas, np.array([coords]), False, line_colour, line_width)
    
        if flags.label:
            if isinstance(coords[0], list):
                label_coords = tuple(coords[0])
            elif isinstance(coords[0], np.ndarray):
                label_coords = tuple(coords[0][0])
            elif isinstance(coords[0], tuple):
                label_coords = coords[0]
    
            cv2.putText(
                canvas,
                label,
                label_coords,
                cv2.FONT_HERSHEY_SIMPLEX,
                label_size,
                label_colour,
                label_width,
                cv2.LINE_AA,
            )
    
    if fill > 0:
        # Blend the original canvas with the filled polygons
        original_canvas = copy.deepcopy(image)
        canvas = cv2.addWeighted(original_canvas, 1 - fill, canvas, fill, 0)

    return canvas
        


def draw_polyline(
    image, 
    annotations, 
    line_colour="default", 
    line_width="auto", 
    show_nodes=False,
    node_colour="default",
    node_size="auto",   
    **kwargs
):
    """
    Draw masks into an image. This function is also used to draw the perimeter 
    of a created or detected reference card.
    
    Parameters
    ----------        
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing lines
    line_colour: {"default", ... see phenopype.print_colours()} str, optional
        contour line colour - default colour as specified in settings
    line_width: {"auto", ... int > 0} int, optional 
        contour line width - automatically scaled to image by default
        DESCRIPTION. The default is "auto".
    show_nodes : bool, optional
        show nodes of polyline. The default is False.
    node_colour : str, optional
        colour of node points. The default is "default".
    node_size : int, optional
        size of node points. The default is "auto".

    Returns
    -------
    image: ndarray
        canvas with lines
    """

    
    # =============================================================================
    # setup

    show_nodes = kwargs.get("show_nodes", False)
    node_colour = kwargs.get("node_colour", "default")
    node_size = kwargs.get("node_size", "auto")

    ## flags
    flags = make_dataclass(cls_name="flags", 
                           fields=[("show_nodes", bool, show_nodes)])

    # =============================================================================
    # annotation management

    annotation_type = _vars._line_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    lines = annotation["data"][annotation_type]

    # =============================================================================
    # setup

    
    line_width = ul._get_size(image.shape[1], image.shape[0], "line_width", line_width)
    line_colour = ul._get_bgr(line_colour, "line_colour")
    if flags.show_nodes:
        node_size = ul._get_size(image.shape[1], image.shape[0], "point_size", node_size)
        node_colour = ul._get_bgr(node_colour, "line_colour")


    # =============================================================================
    # execute

    canvas = copy.deepcopy(image)

    ## draw lines
    for coords in lines:
        cv2.polylines(
            canvas, 
            np.array([coords]), 
            False, 
            line_colour, 
            line_width
            )
        if flags.show_nodes:
            for node in coords:
                print(node)
                cv2.circle(
                    canvas,
                    tuple(node),
                    node_size,
                    node_colour,
                    -1
                    )
                   
                   
    # =============================================================================
    # return

    return canvas


def draw_QRcode(
    image,
    annotations,
    line_colour="default",
    line_width="auto",
    label=True,
    label_colour="default",
    label_size="auto",
    label_width="auto",
    **kwargs,
):   
    """
    

    Parameters
    ----------
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing QR-code (comment)
    line_colour: {"default", ... see phenopype.print_colours()} str, optional
        contour line colour - default colour as specified in settings
    line_width: {"auto", ... int > 0} int, optional 
        contour line width - automatically scaled to image by default
    label : bool, optional
        draw reference label
    label_colour : {"default", ... see phenopype.print_colours()} str, optional
        contour label colour - default colour as specified in settings
    label_size: {"auto", ... int > 0} int, optional 
        contour label font size - automatically scaled to image by default
    label_width:  {"auto", ... int > 0} int, optional 
        contour label font thickness - automatically scaled to image by default
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    canvas : TYPE
        DESCRIPTION.

    """
    # =============================================================================
    # setup

    flags = make_dataclass(cls_name="flags", fields=[("label", bool, label)])
        
    line_width = ul._get_size(image.shape[1], image.shape[0], "line_width", line_width)
    label_size = ul._get_size(image.shape[1], image.shape[0], "label_size", label_size)
    label_width = ul._get_size(image.shape[1], image.shape[0], "label_width", label_width)
    
    line_colour = ul._get_bgr(line_colour, "line_colour")
    label_colour = ul._get_bgr(label_colour, "label_colour")

    # =============================================================================
    # annotation management

    annotation_type = _vars._comment_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )
        
    points = annotation["data"][_vars._mask_type]
    label = annotation["data"][annotation_type]

    canvas = cv2.polylines(
        image, 
        [np.asarray(points, np.int32)], 
        True, 
        line_colour, 
        line_width)

    if flags.label:
        
        (x,y), _ = cv2.minEnclosingCircle(np.array(points))

        label_coords = (int(x),int(y))

        cv2.putText(
            canvas,
            label,
            label_coords,
            cv2.FONT_HERSHEY_SIMPLEX,
            label_size,
            label_colour,
            label_width,
            cv2.LINE_AA,
        )
                

    # =============================================================================
    # return

    return canvas

def draw_reference(
    image,
    annotations,
    line_colour="default",
    line_width="auto",
    label=True,
    label_colour="default",
    label_size="auto",
    label_width="auto",
    **kwargs,
):
    """
    

    Parameters
    ----------
    image : ndarray
        image used as canvas 
    annotation: dict
        phenopype annotation containing reference data
    line_colour: {"default", ... see phenopype.print_colours()} str, optional
        contour line colour - default colour as specified in settings
    line_width: {"auto", ... int > 0} int, optional 
        contour line width - automatically scaled to image by default
    label : bool, optional
        draw reference label
    label_colour : {"default", ... see phenopype.print_colours()} str, optional
        contour label colour - default colour as specified in settings
    label_size: {"auto", ... int > 0} int, optional 
        contour label font size - automatically scaled to image by default
    label_width:  {"auto", ... int > 0} int, optional 
        contour label font thickness - automatically scaled to image by default

    Returns
    -------
    canvas : TYPE
        DESCRIPTION.

    """

    # =============================================================================
    # annotation management

    annotation_type = _vars._reference_type
    annotation_id = kwargs.get(annotation_type + "_id", None)

    annotation = ul._get_annotation(
        annotations=annotations,
        annotation_type=annotation_type,
        annotation_id=annotation_id,
        kwargs=kwargs,
    )

    px_ratio, unit = annotation["data"][annotation_type]
    polygons = annotation["data"][_vars._mask_type]

    # =============================================================================
    # setup

    ## flags
    flags = make_dataclass(cls_name="flags", fields=[("label", bool, label)])
    
    line_width = ul._get_size(image.shape[1], image.shape[0], "line_width", line_width)
    label_size = ul._get_size(image.shape[1], image.shape[0], "label_size", label_size)
    label_width = ul._get_size(image.shape[1], image.shape[0], "label_width", label_width)
    
    line_colour = ul._get_bgr(line_colour, "line_colour")
    label_colour = ul._get_bgr(label_colour, "label_colour")

    # =============================================================================
    # execute

    canvas = copy.deepcopy(image)

    ## draw referenc mask outline
    if len(annotation["data"][_vars._mask_type]) > 0:
        mask_coord_list = annotation["data"][_vars._mask_type]
        cv2.polylines(canvas, np.array([mask_coord_list[0]]), False, line_colour, line_width)
    
    if "support" in annotation["data"]:
        point_coord_list = annotation["data"]["support"]
        cv2.polylines(canvas, np.array([point_coord_list]), False, line_colour, line_width)


    ## draw scale
    if flags.label:
        height, width = canvas.shape[:2]

        hp, wp = height / 100, width / 100

        length = int(px_ratio * 10)

        scale_box = [
            (int(wp * 3), int(hp * 3)),
            (int((wp * 3) + length + (wp * 4)), int(hp * 3)),
            (int((wp * 3) + length + (wp * 4)), int(hp * 11)),
            (int(wp * 3), int(hp * 11)),
            (int(wp * 3), int(hp * 3)),
        ]

        cv2.fillPoly(
            canvas, np.array([scale_box]), ul._get_bgr("lightgrey")
        )

        scale_box_inner = [
            (int(wp * 5), int(hp * 5)),
            (int((wp * 5) + length), int(hp * 5)),
            (int((wp * 5) + length), int(hp * 9)),
            (int(wp * 5), int(hp * 9)),
            (int(wp * 5), int(hp * 5)),
        ]

        cv2.fillPoly(canvas, np.array([scale_box_inner]), line_colour)

        cv2.polylines(
            canvas,
            np.array([scale_box_inner]),
            False,
            line_colour,
            line_width,
        )

        cv2.putText(
            canvas,
            "10 " + unit,
            (int(wp * 6), int(wp * 5)),
            cv2.FONT_HERSHEY_SIMPLEX,
            label_size,
            label_colour,
            label_width * 2,
            cv2.LINE_AA,
        )

    # =============================================================================
    # return

    return canvas


def select_canvas(image, canvas="raw", multi_channel=True, invert=False, **kwargs):
    """
    Isolate a colour channel from an image or select canvas for the pype method.

    Parameters
    ----------
    image : ndarray
        image used as canvas 
    canvas : {"mod", "bin", "gray", "raw", "red", "green", "blue"} str, optional
        the type of canvas to be used for visual feedback. some types require a
        function to be run first, e.g. "bin" needs a segmentation algorithm to be
        run first. black/white images don't have colour channels. coerced to 3D
        array by default
    multi: bool, optional
        coerce returned array to multichannel (3-channel)

    Returns
    -------
    canvas : ndarray
        canvas for drawing

    """

    if image.__class__.__name__ == "_Container":
        
        ## method
        if canvas == "mod":
            image.canvas = copy.deepcopy(image.image)
            ul._print("- modifed image")
        elif canvas == "raw":
            image.canvas = copy.deepcopy(image.image_copy)
            ul._print("- raw image")
        elif canvas == "bin":
            image.canvas = copy.deepcopy(image.image_bin)
            ul._print("- binary image")
        elif canvas == "gray":
            image.canvas = cv2.cvtColor(image.image_copy, cv2.COLOR_BGR2GRAY)
            ul._print("- grayscale image")
        elif canvas == "blue":
            image.canvas = image.image_copy[:, :, 0]
            ul._print("- blue channel")
        elif canvas == "green":
            image.canvas = image.image_copy[:, :, 1]
            ul._print("- green channel")
        elif canvas == "red":
            image.canvas = image.image_copy[:, :, 2]
            ul._print("- red channel")
        else:
            ul._print("- invalid selection - defaulting to raw image", lvl=2)
            image.canvas = copy.deepcopy(image.image_copy)
            
        if invert == True:
            image.canvas = cv2.bitwise_not(image.canvas)
            ul._print("- inverted image")


    elif image.__class__.__name__ == "ndarray":
        if canvas == "raw":
            canvas = copy.deepcopy(image)
            ul._print("- raw image")
        elif canvas == "gray":
            canvas = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            ul._print("- grayscale image")
        elif canvas == "blue":
            canvas = image[:, :, 0]
            ul._print("- blue channel")
        elif canvas == "green":
            canvas = image[:, :, 1]
            ul._print("- green channel")
        elif canvas == "red":
            canvas = image[:, :, 2]
            ul._print("- red channel")  
        else:
            canvas = copy.deepcopy(image)
            ul._print("- invalid selection - defaulting to raw image", lvl=2)
            
        if invert == True:
            canvas = cv2.bitwise_not(canvas)
            ul._print("- inverted image")

    ## check if colour
    if multi_channel:
        if image.__class__.__name__ == "_Container":
            if len(image.canvas.shape) < 3:
                image.canvas = cv2.cvtColor(image.canvas, cv2.COLOR_GRAY2BGR)
        elif image.__class__.__name__ == "ndarray":
            if len(canvas.shape) < 3:
                canvas = cv2.cvtColor(canvas, cv2.COLOR_GRAY2BGR)
                
    return canvas
